import time
import uuid
from enum import Enum
import argparse
from pathlib import Path
from PIL import Image
from mlagents_envs.environment import UnityEnvironment
from gym import spaces
from gym_unity.envs import UnityToGymWrapper
from mlagents_envs.side_channel.engine_configuration_channel import EngineConfigurationChannel
from mlagents_envs.side_channel.environment_parameters_channel import EnvironmentParametersChannel
from mlagents_envs.side_channel.float_properties_channel import FloatPropertiesChannel
from ..env import MapSideChannel

timerEntries = []


class ObservationMode(Enum):
    Vector = 0
    Visual = 1
    VectorVisual = 2


class RLTask(Enum):
    PointNav = 0
    SimpleObjectNav = 1
    ObjectNav = 2


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('binary', metavar='B', help="File path to Unity game binary")
    parser.add_argument('-t', default="PointNav", choices=["PointNav", "SimpleObjectNav", "ObjectNav"],
                        help="Navigational Task that agent will train")
    parser.add_argument('-a', default="Vector", choices=["Vector", "Visual", "VectorVisual"], help="Type of agent")
    parser.add_argument('-s', default=0, type=int, choices=[0, 1, 2], help="Segmentation model for visual observations")
    parser.add_argument('-e', default=2500, type=int, help="Episode length integer")
    parser.add_argument('--worker_id', default=0, type=int,
                        help="Worker Id used to offset comms port for multiple instances")
    parser.add_argument('--seed', default=0, type=int, help="Unity Engine seed")
    parser.add_argument('--timeout', default=1000, type=int,
                        help="Amount of time in seconds to wait for a connection to the binary")
    parser.add_argument('--timescale', default=2, type=int, help="Unity Engine physics timescale modifier")
    parser.add_argument('--quality-level', default=0, type=int, help="Unity Engine quality level")
    parser.add_argument('--nophysics', action="store_true", help="Disable agent car physics simulation")
    parser.add_argument('--saveimages', action="store_true",
                        help="Save images at every visual observations at every step")
    args = parser.parse_args()

    print(args)
    # Debug
    saveImageFlag = args.saveimages
    saveOnStepIfFactorOf = 100

    # TODO bring to a config
    # Rewards
    GOAL_COLLISION_REWARD = 1
    EXP_COLLISION_REWARD = .8
    OTHER_COLLISION_REWARD = -0.9
    FALL_OFF_MAP_REWARD = -0.1
    STEP_REWARD = -0.08

    engine_side_channel = EngineConfigurationChannel()
    environment_side_channel = EnvironmentParametersChannel()
    map_side_channel = MapSideChannel()
    float_properties_side_channel = FloatPropertiesChannel()

    # Connect to Unity Editor environment
    unityEnvironmentStr = str(args.binary) if args.binary else None
    # Connect to specified binary environment
    unity_env = UnityEnvironment(file_name=unityEnvironmentStr, worker_id=args.worker_id, seed=args.seed,
                                 timeout_wait=args.timeout,
                                 side_channels=[engine_side_channel, environment_side_channel, map_side_channel, float_properties_side_channel])

    # Engine Side Channels
    engine_side_channel.set_configuration_parameters(time_scale=args.timescale, quality_level=args.quality_level)

    # Rewards
    environment_side_channel.set_float_parameter("rewardForGoalCollision", GOAL_COLLISION_REWARD)
    environment_side_channel.set_float_parameter("rewardForExplorationPointCollision", EXP_COLLISION_REWARD)
    environment_side_channel.set_float_parameter("rewardForOtherCollision", OTHER_COLLISION_REWARD)
    environment_side_channel.set_float_parameter("rewardForFallingOffMap", FALL_OFF_MAP_REWARD)
    environment_side_channel.set_float_parameter("rewardForEachStep", STEP_REWARD)

    # Segmentation
    # Object Segmentation 0
    # Tag Segmentation 1
    # Layer Segmentation 2
    segmentationMode = args.s
    environment_side_channel.set_float_parameter("segmentationMode", segmentationMode)

    # Agent Selection
    observationModeIndex = ObservationMode[args.a].value
    environment_side_channel.set_float_parameter("observationMode", observationModeIndex)

    # Agent Car Physics
    # if flag is set, hence true, then set agentCarPhysics to 0, false then 1
    agentCarPhysics = int(not args.nophysics)
    environment_side_channel.set_float_parameter("agentCarPhysics", agentCarPhysics)

    # Episode Length
    episodeLength = args.e
    environment_side_channel.set_float_parameter("episodeLength", episodeLength)

    # Task Selection
    selectedTaskIndex = RLTask[args.t].value
    environment_side_channel.set_float_parameter("selectedTaskIndex", selectedTaskIndex)

    # Goal Selection
    # 0 - Tocus;
    # 1 - sedan1
    # 2 - Car1
    # 3 - Car2
    # 4 - City Bus
    # 5 - Sporty_Hatchback
    # Else - SEDAN
    # goalSelectionIndex = 0
    # environment_side_channel.set_float_parameter("goalSelectionIndex", goalSelectionIndex)

    # File to save step time benchmarks
    BASEFILENAME = "bench" + uuid.uuid4().hex + "-a" + args.a + "-t" + str(args.t) + "-s" + str(
        segmentationMode) + "-e" + str(episodeLength) + "-p" + str(agentCarPhysics) + "-sv" + str(int(saveImageFlag))
    print(BASEFILENAME)
    timestr = time.strftime("%Y%m%d-%H%M%S")
    baseFileNameWithTime = BASEFILENAME + "-" + timestr
    csvFilename = baseFileNameWithTime + ".csv"
    # Visual Observation Save outs
    if (saveImageFlag and (observationModeIndex == 1 or observationModeIndex == 2)):
        imgBasePath = Path("") / Path(baseFileNameWithTime)
        imgBasePath.mkdir(parents=True, exist_ok=True)

    # Create Gym Environment
    gym_env = UnityToGymWrapper(unity_env, False, False, True)
    observation = gym_env.reset()

    print('=================================')
    print(gym_env.action_space)
    print(gym_env.action_space.sample())
    print(type(gym_env.action_space), gym_env.action_space.sample(), type(gym_env.action_space.sample()))
    print(gym_env.observation_space)
    print(observation)
    print(gym_env.metadata)
    print(gym_env.spec)
    print(gym_env.name)
    print(gym_env._env.get_steps(gym_env.name))
    print(gym_env.action_space.high, '+++++', gym_env.action_space.low)
    print('=================================')

    start_time = time.time()
    last_time = start_time

    # stepTimeDeltas format: step#, steptimedelta
    stepTimeDeltas = []
    for t in range(30001):
        # action = gym_env.action_space.sample()
        # some arbitrary turns and then reverse
        if (t >= 0):
            action = [1, 1, 0]
        if (t > 17000):
            action = [1, 0, 0]
        if (t > 25000):
            action = [0, 1, 0]
        observation, reward, done, info = gym_env.step(action)

        current_time = time.time()
        stepTimeDeltas.append("{},{}".format(t, current_time - last_time))
        last_time = current_time
        if (saveImageFlag and t % saveOnStepIfFactorOf == 0 and (
                observationModeIndex == 1 or observationModeIndex == 2)):
            # calculateStepTime(start_time, last_time)
            # print (observation)
            storeObservationImage(imgBasePath, observation[0], False, t + 0)
            storeObservationImage(imgBasePath, observation[1], True, t + 1)
            storeObservationImage(imgBasePath, observation[2], False, t + 2)
            # last_time = time.time()
        if done:
            print("\tEpisode finished after {} timesteps".format(t + 1))

    with open(csvFilename, "w") as logfile:
        for s in stepTimeDeltas:
            logfile.write(s + "\n")

    gym_env.close()


def calculateStepTime(start_time, last_time):
    current_time = time.time()
    print("--- %s seconds ---" % (current_time - start_time))
    timerEntries.append(10000 / (current_time - last_time))
    print("Steps/second = {}".format(timerEntries[-1]))
    averageStepTime = 0
    for j in timerEntries:
        averageStepTime += j
    print("Average Steps/second = {}".format(averageStepTime / len(timerEntries)))


def storeObservationImage(basepath, observation, grayscale, observationId):
    # print (observation)
    if (grayscale):
        img = Image.fromarray((observation * 255).astype('uint8').reshape(observation.shape[0], observation.shape[1]),
                              'L')
    else:
        img = Image.fromarray((observation * 255).astype('uint8'), 'RGB')
    img_path = basepath / f"visual_observation{observationId}.png"
    img.save(img_path)
    # img.show()

# For python debugger to directly run this script
if __name__ == "__main__":
    main()



import time
import argparse
from pathlib import Path
from PIL import Image
from mlagents_envs.environment import UnityEnvironment
from gym import spaces
from gym_unity.envs import UnityToGymWrapper
from mlagents_envs.side_channel.engine_configuration_channel import EngineConfigurationChannel
from mlagents_envs.side_channel.environment_parameters_channel import EnvironmentParametersChannel
import socket

def main(args : argparse.Namespace):

    episodes = 50
    for unitySeed in [1,123,999]:
        for episode_sequence in range(10):
            #Rewards
            GOAL_COLLISION_REWARD = 1
            EXP_COLLISION_REWARD = .8
            OTHER_COLLISION_REWARD = -0.9
            FALL_OFF_MAP_REWARD = -0.1
            STEP_REWARD = -0.08

            MAX_STEPS=5

            engine_side_channel = EngineConfigurationChannel()
            environment_side_channel = EnvironmentParametersChannel()

            #Connect to Unity Editor environment
            unityEnvironmentStr = None
            #Connect to specified binary environment
            #unityEnvironmentStr = "../envs/Berlin_v2/Berlin_Walk_v2.exe"
            unityEnvironmentStr = "/data/work/unity-envs/Build2.6.3/Berlin_Walk_V2"
            unity_env = UnityEnvironment(file_name = unityEnvironmentStr, seed = unitySeed, timeout_wait=1000, side_channels =[engine_side_channel, environment_side_channel])

            #Engine Side Channels
            engine_side_channel.set_configuration_parameters(time_scale=1, quality_level = 0)

            #Rewards
            environment_side_channel.set_float_parameter("rewardForGoalCollision", GOAL_COLLISION_REWARD)
            environment_side_channel.set_float_parameter("rewardForExplorationPointCollision", EXP_COLLISION_REWARD)
            environment_side_channel.set_float_parameter("rewardForOtherCollision", OTHER_COLLISION_REWARD)
            environment_side_channel.set_float_parameter("rewardForFallingOffMap", FALL_OFF_MAP_REWARD)
            environment_side_channel.set_float_parameter("rewardForEachStep", STEP_REWARD)

            environment_side_channel.set_float_parameter("observationMode",0) # vector agent
            environment_side_channel.set_float_parameter("episodeLength", MAX_STEPS+2)
            environment_side_channel.set_float_parameter("selectedTaskIndex", 0) # pointnav

            #File to save step time benchmarks
            BASEFILENAME = f"{socket.gethostname()}-seed-{unitySeed}-seq-{episode_sequence}"
            print(BASEFILENAME)
            timestr = time.strftime("%Y%m%d-%H%M%S")
            baseFileNameWithTime = BASEFILENAME + "-" + timestr
            csvFilename = BASEFILENAME + ".csv"

            #Create Gym Environment
            gym_env = UnityToGymWrapper(unity_env, False, False, True)

            with open(csvFilename, "w") as logfile:
                for t in range(episodes):
                    print("Starting episode",t)
                    s = gym_env.reset()
                    for i in range(MAX_STEPS):
                        o = gym_env.step(gym_env.action_space.sample())
                    logfile.write(','.join(str(e) for e in s[0])+'\n')
            gym_env.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    args = parser.parse_args()

    main(args)


import json
import yaml
import copy
import numpy as np
import cattr


# from . import log_util
# logger = log_util.get_logger()

def json_or_yaml(filename):
    """
    This function would be obsolete when pyyaml supports yaml 1.2
    With yaml 1.2 pyyaml can also read json files
    :return:
    """
    import re
    from pathlib import Path

    commas = re.compile(r',(?=(?![\"]*[\s\w\?\.\"\!\-\_]*,))(?=(?![^\[]*\]))')
    """
    Find all commas which are standalone 
     - not between quotes - comments, answers
     - not between brackets - lists
    """
    file_path = Path(filename)
    signs = commas.findall(file_path.open('r').read())
    return "json" if len(signs) > 0 else "yaml"


def load_dict_from_json_file(filename):
    dict_obj = json.load(open(filename, 'r'))
    return dict_obj

def load_dict_from_yaml_file(filename):
    dict_obj = yaml.safe_load(open(filename, 'r'))
    return dict_obj

def load_dict_from_json_str(string):
    dict_obj = json.loads(string)
    return dict_obj


class NPJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NPJSONEncoder, self).default(obj)


def save_to_json_file(obj, filename, sort_keys=False, indent=4):
    if not isinstance(obj, dict):
        obj = obj.__dict__
    json.dump(obj, open(filename, 'w'), indent=indent, sort_keys=sort_keys, cls=NPJSONEncoder)


def save_to_yaml_file(obj, filename):
    if not isinstance(obj, dict):
        obj = obj.__dict__
    yaml.safe_dump(obj, open(filename, 'w'))


class ObjDict(dict):
    """A data structure that inherits from dict and adds object style member access

    TODO:
        * Create an init method that converts nested dicts in this object.
    """
    #__getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    #def __str__(self):
    #    return self.json_dumps()

    def to_json(self, sort_keys=False, indent=2):
        """convert to json

        Args:
            sort_keys: Sort the keys of dict or not, default False
            indent: indentation for JSON struct, default 2 spaces

        Returns:
            json string

        """
        return json.dumps(self, indent=indent, sort_keys=sort_keys, cls=NPJSONEncoder)

    def to_dict(self):
        """convert to dict

        Returns:
            dict object
        """
        return cattr.unstructure(self)

    def to_yaml(self):
        """convery to yaml representation

        Returns:
            yaml string
        """
        return yaml.dump(json.loads(json.dumps(self)))

    def save_to_json_file(self, filename, sort_keys=False, indent=2):
        """Save to json file

        Args:
            filename: path or name of the file
            sort_keys: whether to sort the keys
            indent: indentation of the spaces

        Returns:

        """
        save_to_json_file(self, filename, sort_keys, indent)
        return self

    def save_to_yaml_file(self, filename):
        """Save to yaml file

        Args:
            filename: path or name of the file

        Returns:

        """
        save_to_yaml_file(self, filename)
        return self

    @staticmethod
    def load_from_file(filename):
        """load objdict from a file

        Args:
            filename: path or name of the file

        Returns:
            ObjDict object
        """
        filetype = json_or_yaml(filename)
        if filetype == "json":
            obj = ObjDict.load_from_json_file(filename)
        elif filetype == "yaml":
            obj = ObjDict.load_from_yaml_file(filename)
        return obj

    @staticmethod
    def load_from_json_file(filename):
        """load objdict from a file

        Args:
            filename: path or name of the file

        Returns:
            ObjDict object
        """
        obj = ObjDict(load_dict_from_json_file(filename))
        return obj

    @staticmethod
    def load_from_yaml_file(filename):
        """load objdict from a file

        Args:
            filename: path or name of the file

        Returns:
            ObjDict object
        """
        obj = ObjDict(load_dict_from_yaml_file(filename))
        return obj

    def deepcopy(self):
        """Make a deep copy of itself

        Returns:
            ObjDict object
        """
        return copy.deepcopy(self)



import csv
import struct
from pathlib import Path
import numpy as np
import uuid

from typing import Any, List, Union

import time
import math

from gym_unity.envs import (
    UnityToGymWrapper,
    GymStepResult
)

from mlagents_envs.logging_util import get_logger

from mlagents_envs.exception import UnityWorkerInUseException
from mlagents_envs.environment import UnityEnvironment
from mlagents_envs.side_channel.side_channel import (
    SideChannel,
    IncomingMessage,
    OutgoingMessage,
)
from mlagents_envs.side_channel.engine_configuration_channel import \
    EngineConfigurationChannel
from mlagents_envs.side_channel.environment_parameters_channel import \
    EnvironmentParametersChannel

from mlagents_envs.side_channel.float_properties_channel import \
    FloatPropertiesChannel

try:
    from cv2 import imwrite as imwrite

    print("Navsim: using cv2 as image library")
except:
    try:
        from imageio import imwrite as imwrite

        print("Navsim: using imageio as image library")
    except:
        try:
            from matplotlib.pyplot import imsave as imwrite

            print("Navsim: using matplotlib as image library")
        except:
            try:
                from PIL import Image

                print("Navsim: using PIL as image library")


                def imwrite(filename, arr):
                    im = Image.fromarray(arr)
                    im.save(filename)
            except:
                def imwrite(filename=None, arr=None):
                    print("Navsim: unable to load any of the following "
                          "image libraries: cv2, imageio, matplotlib, "
                          "PIL. Install one of these libraries to "
                          "save visuals.")


                imwrite()


def navsimgymenv_creator(env_config):
    return NavSimGymEnv(env_config)  # return an env instance


class NavSimGymEnv(UnityToGymWrapper):
    """NavSimGymEnv Class is a wrapper to Unity2Gym that inherits from the Gym interface

    Read the **NavSim Environment Tutorial** on how to use this class.
    """
    metadata = {
        'render.modes': ['rgb_array', 'depth', 'segmentation', 'vector']}
    logger = get_logger(__name__)

    def __init__(self, env_config) -> None:
        """
        env_config: The environment configuration dictionary or ObjDict Object
        """

        # TODO: convert env_config to self.env_config so we can add missing values
        #   and use self.env_config to print in the info section
        # filename: Optional[str] = None, obs_mode: int = 0, max_steps:int = 5):
        self.env_config = env_config
        self.obs_mode = int(self.env_config.get('obs_mode', 2))
        self.start_from_episode = int(
            self.env_config.get('start_from_episode', 1))
        self.debug = env_config.get("debug", False)
        self.obs = None

        if self.obs_mode == 0:
            env_config["save_visual_obs"] = False
        elif self.obs_mode == 1:
            env_config["save_vector_obs"] = False

        self.save_actions = env_config.get("save_actions", False)
        self.save_visual_obs = env_config.get("save_visual_obs", False)
        self.save_vector_obs = env_config.get("save_vector_obs", False)

        self.e_num = 0
        self.s_num = 0

        self._agent_position = None
        self._agent_velocity = None
        self._agent_rotation = None
        self._goal_position = None

        self.spl_start = self.spl_current = None
        self.reward_spl_delta_mul = float(
            self.env_config.get('reward_spl_delta_mul', 1))

        # self.run_base_folder_str = env_config.get("run_base_folder_str", '.')
        # self.run_base_folder = Path(self.run_base_folder_str)
        seed = int(self.env_config.get('seed') or 0)

        # if self._env:
        #    raise ValueError('Environment already open')
        # else:

        self.map_side_channel = MapSideChannel()
        self.fpc = FloatPropertiesChannel()

        # print(self.map_side_channel)

        eng_sc = EngineConfigurationChannel()
        eng_sc.set_configuration_parameters(time_scale=10, quality_level=0)

        env_pc = EnvironmentParametersChannel()
        env_sfp = env_pc.set_float_parameter

        env_sfp("rewardForGoal",
                float(self.env_config.get('reward_for_goal', 50)))
        env_sfp("rewardForNoViablePath",
                float(self.env_config.get('reward_for_no_viable_path', -50)))
        env_sfp("rewardStepMul",
                float(self.env_config.get('reward_step_mul', 0.1)))
        env_sfp("rewardCollisionMul",
                float(self.env_config.get('reward_collision_mul', 4)))
        env_sfp("rewardSplDeltaMul",
                float(self.env_config.get('reward_spl_delta_mul', 1)))
        env_sfp("segmentationMode",
                float(self.env_config.get('segmentation_mode', 1)))
        env_sfp("observationMode",
                float(self.env_config.get('obs_mode', 2)))
        env_sfp("episodeLength",
                float(self.env_config.get('episode_max_steps', 100)))
        env_sfp("selectedTaskIndex", float(self.env_config.get('task', 0)))
        env_sfp("goalSelectionIndex", float(self.env_config.get('goal', 0)))
        env_sfp("agentCarPhysics",
                float(self.env_config.get('agent_car_physics', 0)))
        env_sfp("goalDistance", float(self.env_config.get('goal_distance', 10)))
        env_sfp("numberOfTrafficVehicles",
                float(self.env_config.get('traffic_vehicles', 0)))

        env_path = self.env_config.get('env_path')
        env_path = None if env_path is None else str(Path(env_path).resolve())

        self._navsim_base_port = self.env_config.get('base_port', None)
        if self._navsim_base_port is None:
            self._navsim_base_port = UnityEnvironment.BASE_ENVIRONMENT_PORT if env_path else UnityEnvironment.DEFAULT_EDITOR_PORT
        self._navsim_worker_id = self.env_config.get('worker_id', 0)

        while True:
            try:
                log_folder = Path(
                    self.env_config.get('log_folder', './env_log')).resolve()
                log_folder.mkdir(parents=True, exist_ok=True)
                ad_args = [
                    "-force-device-index",
                    f"{self.env_config.get('env_gpu_id', 0)}",
                    "-observationWidth",
                    f"{self.env_config.get('obs_width', 256)}",
                    "-observationHeight",
                    f"{self.env_config.get('obs_height', 256)}",
                    "-fastForward", f"{self.start_from_episode - 1}",
                    "-showVisualObservations" if self.env_config.get(
                        'show_visual', False) else "",
                    "-saveStepLog" if self.debug else ""
                ]
                self.uenv = UnityEnvironment(file_name=env_path,
                                             log_folder=str(log_folder),
                                             seed=seed,
                                             timeout_wait=self.env_config.get(
                                                 'timeout', 600) + (0.5 * (
                                                     self.start_from_episode - 1)),
                                             worker_id=self._navsim_worker_id,
                                             base_port=self._navsim_base_port,
                                             no_graphics=False,
                                             side_channels=[eng_sc, env_pc,
                                                            self.map_side_channel,
                                                            self.fpc],
                                             additional_args=ad_args)
            except UnityWorkerInUseException:
                time.sleep(2)
                self._navsim_base_port += 1
            else:
                NavSimGymEnv.logger.info(f"Created UnityEnvironment at port "
                                         f"{self._navsim_base_port + self._navsim_worker_id}"
                                         f" to start from episode {self.start_from_episode}")
                break

        super().__init__(unity_env=self.uenv,
                         uint8_visual=False,
                         flatten_branched=False,
                         allow_multiple_obs=True,
                         action_space_seed=seed
                         )

        # TODO: Once the environment has capability to start from an episode, then remove this
        # if self.start_from_episode > 1:
        #    logger.info(f'jumping to episode {self.start_from_episode}')
        # for i in range(1, self.start_from_episode):
        #    self.reset()
        #    logger.info(f'skipping episode {self.e_num}')

        # TODO: the filenames should be prefixed with specific id of this instance of env

        # TODO: Read the file upto start_episode and purge the records
        self.actions_file = log_folder / 'actions.csv'
        if self.save_actions:
            if (self.start_from_episode == 1) or (
                    self.actions_file.exists() == False):
                self.actions_file = self.actions_file.open(mode='w')
            else:
                self.actions_file = self.actions_file.open(mode='a')
            self.actions_writer = csv.writer(self.actions_file,
                                             delimiter=',',
                                             quotechar='"',
                                             quoting=csv.QUOTE_MINIMAL)

        if self.save_visual_obs and (self.obs_mode in [1, 2]):
            self.rgb_folder = log_folder / 'rgb_obs'
            self.rgb_folder.mkdir(parents=True, exist_ok=True)
            self.dep_folder = log_folder / 'dep_obs'
            self.dep_folder.mkdir(parents=True, exist_ok=True)
            self.seg_folder = log_folder / 'seg_obs'
            self.seg_folder.mkdir(parents=True, exist_ok=True)
        else:
            self.save_visual_obs = False

        if self.save_vector_obs and (self.obs_mode in [0, 2]):
            self.vec_file = log_folder / 'vec_obs.csv'
            if (self.start_from_episode == 1) or (
                    self.vec_file.exists() == False):
                self.vec_file = self.vec_file.open(mode='w')
                self.vec_writer = csv.writer(self.vec_file,
                                             delimiter=',',
                                             quotechar='"',
                                             quoting=csv.QUOTE_MINIMAL)
                self.vec_writer.writerow(
                    ['e_num', 's_num', 'spl_current', 'timestamp'] +
                    ['posx', 'posy', 'posz', 'velx', 'vely', 'velz',
                     'rotx', 'roty', 'rotz', 'rotw', 'goalx', 'goaly', 'goalz',
                     'proxforward', 'prox45left', 'prox45right'])
            else:
                # TODO: Read the file upto start_episode and purge the records
                self.vec_file = self.vec_file.open(mode='a')
                self.vec_writer = csv.writer(self.vec_file,
                                             delimiter=',',
                                             quotechar='"',
                                             quoting=csv.QUOTE_MINIMAL)
            self.vec_file.flush()
        else:
            self.save_vector_obs = False

    def _save_obs(self, obs):
        if self.save_vector_obs:
            self.vec_writer.writerow(
                [self.e_num, self.s_num, self.spl_current, time.time()] +
                list(obs[-1]))
            self.vec_file.flush()
        if self.save_visual_obs:
            filename = f'{self.e_num}_{self.s_num}.jpg'
            imwrite(str(self.rgb_folder / filename), obs[0] * 255)
            imwrite(str(self.dep_folder / filename), obs[1] * 255)
            imwrite(str(self.seg_folder / filename), obs[2] * 255)

    def reset(self) -> Union[List[np.ndarray], np.ndarray]:
        s0 = super().reset()
        self.obs = s0
        if self.obs_mode in [0, 2]:
            vec_obs = self.obs[-1]
            self._agent_position = vec_obs[0:3]
            self._agent_velocity = vec_obs[3:6]
            self._agent_rotation = vec_obs[6:10]
            self._goal_position = vec_obs[10:13]
        self.s_num = 0
        self.e_num += 1 if self.e_num else self.start_from_episode
        self.spl_start = self.spl_current = self.shortest_path_length
        self._save_obs(self.obs)
        return s0

    def step(self, action: List[Any]) -> GymStepResult:
        s_, r, episode_done, info = super().step(action)
        self.obs = s_
        if self.obs_mode in [0, 2]:
            vec_obs = self.obs[-1]
            self._agent_position = vec_obs[0:3]
            self._agent_velocity = vec_obs[3:6]
            self._agent_rotation = vec_obs[6:10]
            self._goal_position = vec_obs[10:13]
        self.s_num += 1
        self.spl_current = self.shortest_path_length
        self._save_obs(self.obs)
        if self.save_actions:
            self.actions_writer.writerow(
                [self.e_num, self.s_num] + list(action))
            self.actions_file.flush()
        return s_, r, episode_done, info

    # def close(self):
    #    if self.save_vector_obs:
    #        self.vec_file.close()
    #    if self.save_vector_obs or self.save_visual_obs:
    #        self.actions_file.close()
    #    super().close()

    #    @property
    #    def observation_space_shapes(self) -> list:
    #        """Returns the dimensions of the observation space
    #        """
    #        return [obs.shape for obs in self.observation_space.spaces]
    #
    #    @property
    #    def observation_space_types(self) -> list:
    #        """Returns the dimensions of the observation space
    #        """
    #        return [type(obs) for obs in self.observation_space.spaces]

    def render(self, mode='rgb_array') -> None:
        """Returns the image array based on the render mode

        Args:
            mode: 'rgb_array' or 'depth' or 'segmentation' or 'vector'

        Returns:
            For Observation Mode 1 and 2: each render mode returns a numpy array of the image
            For Observation Mode 0 and 2: render mode vector returns vector observations
        """
        if mode == 'rgb_array' and self.obs_mode in [1, 2]:
            obs = self.obs[0]
        elif mode == 'depth' and self.obs_mode in [1, 2]:
            obs = self.obs[1]
        elif mode == 'segmentation' and self.obs_mode in [1, 2]:
            obs = self.obs[2]
        elif mode == 'vector' and self.obs_mode in [0, 2]:
            obs = self.obs[-1]
        else:
            raise ValueError("Bad render mode was specified or the "
                             "observation mode of the environment doesnt "
                             "support this render mode. render mode = "
                             "{mode}, observation mode = {self.obs_mode}")
        return obs

    def get_navigable_map(self, resolution_x=256, resolution_y=256,
                          cell_occupancy_threshold=0.5) -> np.ndarray:
        """Get the Navigable Areas map

        Args:
            resolution_x: The size of the agent_x axis of the resulting grid, default = 256
            resolution_y: The size of the y axis of the resulting grid, default = 256
            cell_occupancy_threshold: If at least this much % of the cell is occupied, then it will be marked as non-navigable, default = 50%

        Returns:
            A numpy array having 0 for non-navigable and 1 for navigable cells,
            and the 0,0 in the top-left corner

        Note:
            Largest resolution is 3284 agent_x 2666
        """
        if (resolution_x > 3284) or (resolution_y > 2666):
            raise ValueError("maximum map size is 3284 agent_x 2666")

        self.uenv._process_immediate_message(
            self.map_side_channel.build_immediate_request("binaryMap",
                                                          [resolution_x,
                                                           resolution_y,
                                                           cell_occupancy_threshold]))
        return self.map_side_channel.requested_map

        # def start_navigable_map(self, resolution_x=256, resolution_y=256,
        #                        cell_occupancy_threshold=0.5):
        """Start the Navigable Areas map

        Args:
            resolution_x: The size of the agent_x axis of the resulting grid, default = 256
            resolution_y: The size of the y axis of the resulting grid, default = 256
            cell_occupancy_threshold: If at least this much % of the cell is occupied, then it will be marked as non-navigable, default = 50%

        Returns:
            Nothing

        Note:
            Largest resolution is 3284 agent_x 2666
        """
        #    if (resolution_x > 3284) or (resolution_y > 2666):
        #        raise ValueError("maximum map size is 3284 agent_x 2666")
        #    self.map_side_channel.send_request("binaryMap",
        #                                       [resolution_x, resolution_y,
        #                                        cell_occupancy_threshold])
        # print('Inside get navigable map function:',self.map_side_channel.requested_map)

        # def get_navigable_map(self) -> np.ndarray:
        """Get the Navigable Areas map

        Args:

        Returns:
            A numpy array having 0 for non-navigable and 1 for navigable cells

        Note:
            This only works if you have called ``reset()`` or ``step()`` on the
            environment at least once after calling start_navigable_map() method.
        """

    #    return self.map_side_channel.requested_map

    def unity_loc_to_navmap_loc(self, unity_x, unity_z, navmap_max_x=256,
                                navmap_max_y=256):
        """

        Args:
            unity_x:
            unity_z:
            navmap_max_x:
            navmap_max_y:

        Returns:
            navmap_x, navmap_y
        """
        unity_max_x, _, unity_max_z = self.unity_map_dims
        #TODO: 0 <= unity_x < math.floor(unity_max_x) && 0 <= unity_z < math.floor(unity_max_z)
        navmap_x = math.floor(unity_x / (math.floor(unity_max_x) / navmap_max_x))
        navmap_y = (navmap_max_y - 1) - math.floor(
            unity_z / (math.floor(unity_max_z) / navmap_max_y))
        return navmap_x, navmap_y

    def navmap_loc_to_unity_loc(self, navmap_x, navmap_y, navmap_max_x=256,
                                navmap_max_y=256, navmap_cell_center=True):
        """

        Args:
            navmap_x:
            navmap_y:
            navmap_max_x:
            navmap_max_y:
            navmap_cell_center:

        Returns:
            unity_x, unity_z
        """
        unity_max_x, unity_max_y, unity_max_z = self.unity_map_dims

        #TODO:  input: 0 <= navmap_x < navmap_max_x && 0<= navmap_y < navmap_max_y
        unity_x = navmap_x * (math.floor(unity_max_x) / navmap_max_x)
        unity_z = math.floor(unity_max_z) - (navmap_y + 1) * (math.floor(unity_max_z) / navmap_max_y)
        if navmap_cell_center:
            unity_x += (math.floor(unity_max_x)/navmap_max_x)/2
            unity_z += (math.floor(unity_max_z)/navmap_max_y)/2

        return unity_x, unity_z

    def sample_navigable_point(self, resolution_x=256, resolution_y=256,
                               cell_occupancy_threshold=0.5):
        """Provides a random sample of navigable point

        Returns: x,y point on the navigable map

        """
        if self.map_side_channel.requested_map is None:
            self.get_navigable_map(resolution_x, resolution_y,
                                   cell_occupancy_threshold)

        idx = np.argwhere(self.map_side_channel.requested_map == 1)
        idx_sample = np.random.choice(len(idx),replace=False)
        return idx[idx_sample]

    @staticmethod
    def register_with_gym():
        """Registers the environment with gym registry with the name navsim

        """
        env_id = 'navsim-v0'
        from gym.envs.registration import register, registry

        env_dict = registry.env_specs.copy()
        for env in env_dict:
            if env_id in env:
                print(f"Navsim: Removing {env} from Gym registry")
                del registry.env_specs[env]

        print(f"Navsim: Adding {env_id} to Gym registry")
        register(id=env_id, entry_point='navsim.env:NavSimGymEnv')

    @staticmethod
    def register_with_ray():
        """Registers the environment with ray registry with the name navsim

        """
        from ray.tune.registry import register_env
        register_env("navsim-v0", navsimgymenv_creator)

    def get_dummy_obs(self):
        # prepare input data
        obs_names = []
        obs_data = []
        for state_dim in [obs.shape for obs in self.observation_space.spaces]:
            # obs_dim = [1] + [state_dim[2],state_dim[1],state_dim[0],
            obs = np.zeros([1] + list(state_dim))
            obs_name = 'state'
            for i in range(len(state_dim)):
                #    input_dim.append(state_dim[i])
                obs_name += f'_{state_dim[i]}'
            obs_names.append(obs_name)
            obs_data.append(obs)

        # print([o.shape for o in obs_data],obs_names)

        return obs_data, obs_names

    def get_dummy_actions(self):
        action_dim = self.action_space.shape[0]
        actions_data = np.zeros([1, action_dim], dtype=np.float)
        actions_names = f'action_{action_dim}'
        return actions_data, actions_names

    # Functions added to have parity with Env and RLEnv of habitat lab
    @property
    def sim(self):
        """Returns an instance of the sim

        Added for compatibility with habitat API.

        Returns: link to self

        """
        return self

    @property
    def unity_map_dims(self):
        """Returns the maximum x,y,z values of Unity Map

        Note: While converting to 2-D map, the Z-axis max of 3-D Unity Map
        corresponds to Y-axis max of 2-D map

        Returns: maximum x, y, z from unity map

        """
        return 3284.0, 52.9, 2666.3

    @property
    def agent_position(self):
        """Position of agent in unity map coordinates x,y,z
        """
        return self._agent_position

    @property
    def agent_velocity(self):
        """Velocity of agent in unity map coordinates x,y,z
        """
        return self._agent_velocity

    @property
    def agent_rotation(self):
        """Rotation of agent in unity map coordinates x,y,z,w (Quaternions)
        """
        return self._agent_rotation

    @property
    def agent_rotation_in_euler(self):
        """Position of agent in Euler coordinates roll_x, pitch_y, yaw_z
        """
        if self.agent_rotation is None:
            return None
        else:
            y, z, x, w = self.agent_rotation
            # Convert a quaternion into euler angles (roll, pitch, yaw)
            # roll is rotation around x in radians (counterclockwise)
            # pitch is rotation around y in radians (counterclockwise)
            # yaw is rotation around z in radians (counterclockwise)

            t0 = +2.0 * (w * x + y * z)
            t1 = +1.0 - 2.0 * (x * x + y * y)
            roll_x = math.atan2(t0, t1)

            t2 = +2.0 * (w * y - z * x)
            t2 = +1.0 if t2 > +1.0 else t2
            t2 = -1.0 if t2 < -1.0 else t2
            pitch_y = math.asin(t2)

            t3 = +2.0 * (w * z + x * y)
            t4 = +1.0 - 2.0 * (y * y + z * z)
            yaw_z = math.atan2(t3, t4)

            return roll_x, pitch_y, yaw_z  # in radians

    @property
    def agent_rotation_in_navmap(self):
        _,_,yaw_z = self.agent_rotation_in_euler
        return yaw_z

    # sim.get_agent_state() -> agent_x, y, orientation
    # sim.set_agent_state(position, orientation)
    # sim.get_observations_at(position, orientation) -> observation when agent is at position with specified orientation
    # sim.sample_navigable_point() -> agent_x,y (must be a navigable location in the map)

    @property
    def agent_rotation_in_navmap(self):
        return None

    @property
    def goal_position(self):
        """Position of goal in unity map coordinates x,y,z
        """
        return self._goal_position

    @property
    def current_episode_num(self):
        """Currently executing episode number, 0 means env just initialized
        """
        return self.e_num

    @property
    def last_step_num(self):
        """Last executed step number, 0 mean env just initialized or reset
        """
        return self.s_num

    @property
    def shortest_path_length(self):
        """Return the Shortest Path Length

        Returns: Return the Shortest Path Length

        """
        return self.fpc.get_property("ShortestPath")


class MapSideChannel(SideChannel):
    """This is the SideChannel for retrieving map data from Unity.
    You can send map requests to Unity using send_request.
    The arguments for a mapRequest are ("binaryMap", [RESOLUTION_X, RESOLUTION_Y, THRESHOLD])
    """
    resolution = None

    def __init__(self) -> None:
        channel_id = uuid.UUID("24b099f1-b184-407c-af72-f3d439950bdb")
        super().__init__(channel_id)
        self.requested_map = None

    def on_message_received(self, msg: IncomingMessage) -> np.ndarray:
        if self.resolution is None:
            return None

        raw_bytes = msg.get_raw_bytes()
        self.requested_map = np.unpackbits(raw_bytes)[
                             0:self.resolution[0] * self.resolution[1]]
        self.requested_map = self.requested_map.reshape((self.resolution[0],
                                                         self.resolution[1]))
        return self.requested_map

    def send_request(self, key: str, value: List[float]) -> None:
        """Sends a request to Unity
        The arguments for a mapRequest are ("binaryMap", [RESOLUTION_X, RESOLUTION_Y, THRESHOLD])
        """
        self.resolution = value
        msg = OutgoingMessage()
        msg.write_string(key)
        msg.write_float32_list(value)
        super().queue_message_to_send(msg)

    def build_immediate_request(self, key: str,
                                value: List[float]) -> bytearray:
        self.resolution = value
        msg = OutgoingMessage()
        msg.write_string(key)
        msg.write_float32_list(value)

        result = bytearray()
        result += self.channel_id.bytes_le
        result += struct.pack("<i", len(msg.buffer))
        result += msg.buffer
        return result



from collections import OrderedDict

import math
import torch
from torch.nn import functional as F

class ModelTorch(torch.nn.Module):
    def __init__(self, conf):
        super().__init__()
        self.conf = conf

        self.fc1 = nn.Linear(self.conf.n_x, 512)
        self.fc2 = nn.Linear(512, 512)
        self.fc3 = nn.Linear(512, self.conf.n_y)

    def forward(self, x):
        """

        :param x: numpy array
        :return:
        """
        x = torch.FloatTensor(x).to(self.device),
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class AirSimNN(object):
    def __init__(self, conf):
        self.conf = conf
        self.model = ModelTorch(conf)
        self.opt_fn = torch.optim.RMSprop(self.model.parameters())
        self.los_fn = torch.nn.MSELoss()
        self.device = torch.device(
            'cpu')  # 'cuda' if torch.cuda.is_available() else 'cpu')

    def fit(self, x_train, y_train, x_valid=None, y_valid=None):
        train_generator = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(
                torch.from_numpy(x_train).type(torch.Tensor).view(
                    [-1, self.conf.n_x]),
                torch.from_numpy(y_train).type(torch.Tensor).view(
                    [-1, self.conf.n_y])
            ), batch_size=self.conf.n_batch_size, shuffle=True
        )
        for epoch in range(self.conf.n_epochs):
            for x_batch, y_batch in train_generator:
                x_batch, y_batch = x_batch.to(self.device), y_batch.to(
                    self.device)
                y_pred = self.model(x_batch)
                loss = self.los_fn(y_pred, y_batch)
                self.opt_fn.zero_grad()
                loss.backward(retain_graph=True)
                self.opt_fn.step()
        return self

    def predict(self, x):
        with torch.no_grad():
            x = torch.from_numpy(x).type(torch.Tensor).view(
                [-1, self.conf.n_x]).to(self.device)
            y_pred = self.model(x)
            y_pred = y_pred.detach().numpy()
        return y_pred

    def info(self):
        print(self.conf)
        print(self.model)


def conv2d_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1):
    h_w, kernel_size, stride, pad, dilation = num2tuple(h_w), \
                                              num2tuple(kernel_size), \
                                              num2tuple(stride), \
                                              num2tuple(pad), \
                                              num2tuple(dilation)
    pad = num2tuple(pad[0]), num2tuple(pad[1])
    # print(h_w,kernel_size,stride,pad,dilation)
    h = math.floor(
        (h_w[0] + sum(pad[0]) - dilation[0] * (kernel_size[0] - 1) - 1) /
        stride[0] + 1)
    w = math.floor(
        (h_w[1] + sum(pad[1]) - dilation[1] * (kernel_size[1] - 1) - 1) /
        stride[1] + 1)

    return h, w


def convtransp2d_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1,
                              out_pad=0):
    h_w, kernel_size, stride, pad, dilation, out_pad = num2tuple(h_w), \
                                                       num2tuple(
                                                           kernel_size), num2tuple(
        stride), num2tuple(
        pad), num2tuple(dilation), num2tuple(out_pad)
    pad = num2tuple(pad[0]), num2tuple(pad[1])

    h = (h_w[0] - 1) * stride[0] - sum(pad[0]) + dilation[0] * (
            kernel_size[0] - 1) + out_pad[0] + 1
    w = (h_w[1] - 1) * stride[1] - sum(pad[1]) + dilation[1] * (
            kernel_size[1] - 1) + out_pad[1] + 1

    return h, w


def conv2d_get_padding(h_w_in, h_w_out, kernel_size=1, stride=1, dilation=1):
    h_w_in, h_w_out, kernel_size, stride, dilation = num2tuple(
        h_w_in), num2tuple(h_w_out), \
                                                     num2tuple(
                                                         kernel_size), num2tuple(
        stride), num2tuple(dilation)

    p_h = ((h_w_out[0] - 1) * stride[0] - h_w_in[0] + dilation[0] * (
            kernel_size[0] - 1) + 1)
    p_w = ((h_w_out[1] - 1) * stride[1] - h_w_in[1] + dilation[1] * (
            kernel_size[1] - 1) + 1)

    return (math.floor(p_h / 2), math.ceil(p_h / 2)), (
        math.floor(p_w / 2), math.ceil(p_w / 2))


def convtransp2d_get_padding(h_w_in, h_w_out, kernel_size=1, stride=1,
                             dilation=1, out_pad=0):
    h_w_in, h_w_out, kernel_size, stride, dilation, out_pad = num2tuple(
        h_w_in), num2tuple(h_w_out), \
                                                              num2tuple(
                                                                  kernel_size), num2tuple(
        stride), num2tuple(
        dilation), num2tuple(out_pad)

    p_h = -(h_w_out[0] - 1 - out_pad[0] - dilation[0] * (kernel_size[0] - 1) - (
            h_w[0] - 1) * stride[0]) / 2
    p_w = -(h_w_out[1] - 1 - out_pad[1] - dilation[1] * (kernel_size[1] - 1) - (
            h_w[1] - 1) * stride[1]) / 2

    return (math.floor(p_h / 2), math.ceil(p_h / 2)), (
        math.floor(p_w / 2), math.ceil(p_w / 2))


class ActorCriticWrapper(torch.nn.Module):
    """Wraps Actor and Critic for the purpose of TensorBoard

    """
    def __init__(self, state_dims, action_dim, max_action):
        super().__init__()
        self.actor = Actor(state_dims, action_dim, max_action)
        self.critic = Critic(state_dims, action_dim)

    def forward(self, state, action):
        q1 = self.actor(state)
        q2 = self.critic(state, action)
        return q1, q2


class Actor(torch.nn.Module):
    """Actor class
    state_dims: Should always be a list of state_dimension
                      if vector is True then vector obs should always be last
                      if visual is True then visual obs should come before vec dimension
    """

    def __init__(self, state_dims, action_dim, max_action):
        super(Actor, self).__init__()

        # All feature_layers end up with ReLU output
        self.feature_layers = torch.nn.ModuleList()
        l_out_size = []

        # visual network settings
        stride = 1
        padding = 0
        dilation = 1
        kernel_size = 2
        out_channels = 4

        # input layer
        for state_dim in state_dims:
            if len(state_dim) == 1:  # means we have vector observation
                out_size = state_dim[0]*16  # make sure its always int
                l_out_size.append(out_size)
                layer = torch.nn.Sequential(OrderedDict([
                    (f'linear_1',torch.nn.Linear(state_dim[0], state_dim[0]*4)),
                    (f'activ_1',torch.nn.ReLU()),
                    (f'linear_2',torch.nn.Linear(state_dim[0]*4, state_dim[0]*8)),
                    (f'activ_2',torch.nn.ReLU()),
                    (f'linear_3',torch.nn.Linear(state_dim[0]*8, state_dim[0]*16)),
                    (f'activ_3',torch.nn.ReLU()),
                    #(f'linear_4',torch.nn.Linear(state_dim[0]*16, state_dim[0]*8)),
                    #(f'activ_4',torch.nn.ReLU()),
                    #(f'linear_5',torch.nn.Linear(state_dim[0]*8, state_dim[0]*4)),
                    #(f'activ_5',torch.nn.ReLU()),
                    #(f'linear_6',torch.nn.Linear(state_dim[0]*4, state_dim[0])),
                    #(f'activ_6',torch.nn.ReLU()),
                ]))
                self.feature_layers.append(layer)

            else:  # visual:
                layer = torch.nn.Sequential(OrderedDict([
                    ('conv_1',torch.nn.Conv2d(in_channels=state_dim[2],
                                    out_channels=out_channels,
                                    kernel_size=kernel_size,
                                    stride=stride,
                                    padding=padding,
                                    dilation=dilation)),
                    # [batch_size, n_features_conv, height, width]
                    ('pool_1',torch.nn.MaxPool2d(kernel_size=kernel_size,
                                       stride=stride,
                                       padding=padding,
                                       dilation=dilation)),
                    ('activ_1',torch.nn.ReLU())
                ]))
                h, w = conv2d_output_shape(h_w=state_dim[0:2],
                                           kernel_size=kernel_size,
                                           stride=stride,
                                           pad=padding,
                                           dilation=dilation)
                h, w = conv2d_output_shape(h_w=[h, w],
                                           kernel_size=kernel_size,
                                           stride=stride,
                                           pad=padding,
                                           dilation=dilation)
                l_out_size.append(h * w * out_channels)
                self.feature_layers.append(layer)

        #        self.l3 = torch.nn.Linear(300, action_dim)

        cat_dim = sum(l_out_size)

        self.out_layer = torch.nn.Sequential(OrderedDict([
            ('linear_out',torch.nn.Linear(cat_dim, action_dim)),
            ('activ_out',torch.nn.Tanh())
        ]))
        # print(f'{l_out_size},cat_dim:{cat_dim},action_dim:{action_dim}')
        # self.action_out = torch.nn.Linear(cat_dim, action_dim)

        self.max_action = max_action

    def forward(self, state):
        """

        :param state: state is a list of numpy/cupy arrays
        :return:
        """
        # curframe = inspect.currentframe()
        # calframe = inspect.getouterframes(curframe, 2)
        # print('actor caller name:', calframe[1][3])
        # print('actor',state.shape)
        # print('actor',self.l1)
        # state = state[0]
        #state = [torch.as_tensor(s,
        #                         dtype=torch.float,
        #                         device=self.device).unsqueeze(0) for s in
        #         state]
        features = []
        for s, layer in zip(state, self.feature_layers):
            layer = layer(s)
            if layer.dim() > 2:
                layer = layer.view(layer.size(0), -1)
            features.append(layer)
            # print(layer.shape)

        #        if len(features) > 1:
        #            features_cat = torch.cat(features, dim=1)
        #        else:
        #            features_

        features_cat = torch.cat(features, dim=1) if len(features) > 1 else \
            features[0]
        a = self.out_layer(features_cat)
        return self.max_action * a


class Critic(torch.nn.Module):
    """Critic class
    TODO:
    Should we combine the state with action after some layers or right at the first layer?
    """

    def __init__(self, state_dims, action_dim):
        super(Critic, self).__init__()

        # All feature_layers end up with ReLU output
        self.feature_layers = torch.nn.ModuleList()
        l_out_size = []

        # visual network settings
        stride = 1
        padding = 0
        dilation = 1
        kernel_size = 2
        out_channels = 4

        for state_dim in state_dims:
            if len(state_dim) == 1:  # means we have vector observation
                out_size = state_dim[0]*16  # make sure its always int
                l_out_size.append(out_size)
                layer = torch.nn.Sequential(OrderedDict([
                    (f'linear_1',torch.nn.Linear(state_dim[0], state_dim[0]*4)),
                    (f'activ_1',torch.nn.ReLU()),
                    (f'linear_2',torch.nn.Linear(state_dim[0]*4, state_dim[0]*8)),
                    (f'activ_2',torch.nn.ReLU()),
                    (f'linear_3',torch.nn.Linear(state_dim[0]*8, state_dim[0]*16)),
                    (f'activ_3',torch.nn.ReLU()),
                    #(f'linear_4',torch.nn.Linear(state_dim[0]*16, state_dim[0]*8)),
                    #(f'activ_4',torch.nn.ReLU()),
                    #(f'linear_5',torch.nn.Linear(state_dim[0]*8, state_dim[0]*4)),
                    #(f'activ_5',torch.nn.ReLU()),
                    #(f'linear_6',torch.nn.Linear(state_dim[0]*4, state_dim[0])),
                    #(f'activ_6',torch.nn.ReLU()),
                ]))
                self.feature_layers.append(layer)
            else:  # visual:
                layer = torch.nn.Sequential(
                    torch.nn.Conv2d(in_channels=state_dim[2],
                                    out_channels=out_channels,
                                    kernel_size=kernel_size,
                                    stride=stride,
                                    padding=padding,
                                    dilation=dilation),
                    # [batch_size, n_features_conv, height, width]
                    torch.nn.MaxPool2d(kernel_size=kernel_size,
                                       stride=stride,
                                       padding=padding,
                                       dilation=dilation),
                    torch.nn.ReLU()
                )
                h, w = conv2d_output_shape(h_w=state_dim[0:2],
                                           kernel_size=kernel_size,
                                           stride=stride,
                                           pad=padding,
                                           dilation=dilation)
                h, w = conv2d_output_shape(h_w=[h, w],
                                           kernel_size=kernel_size,
                                           stride=stride,
                                           pad=padding,
                                           dilation=dilation)
                l_out_size.append(h * w * out_channels)
                self.feature_layers.append(layer)

        # add action also as one of the feature layers
        out_size = action_dim  # make sure its always int
        layer = torch.nn.Sequential(
            torch.nn.Linear(action_dim, out_size),
            torch.nn.ReLU()
            #                torch.nn.Linear(400, 300),
            #                torch.nn.ReLU()
        )
        self.feature_layers.append(layer)
        l_out_size.append(out_size)

        cat_dim = sum(l_out_size)

        self.out_layer = torch.nn.Sequential(
            torch.nn.Linear(cat_dim, 1),
        )

    def forward(self, state, action):
        """

        :param state: a list of torch Tensors
        :param action: a torch Tensor
        :return:
        """
        # print('critic',state.shape)
        # print('critic',self.l1)

        # state = state[0]
        #state = [torch.as_tensor(s,
        #                         dtype=torch.float,
        #                         device=self.device).unsqueeze(0) for s in
        #         state]

        #action = torch.as_tensor(action,
        #                         dtype=torch.float,
        #                         device=self.device)

        features = []
        for s, layer in zip(state, self.feature_layers):
            layer = layer(s)
            if layer.dim() > 2:
                layer = layer.view(layer.size(0), -1)
            features.append(layer)
            # print(layer.shape)

        layer = self.feature_layers[-1]
        layer = layer(action)
        if layer.dim() > 2:
            layer = layer.view(layer.size(0), -1)
        features.append(layer)

        features_cat = torch.cat(features, dim=1) if len(features) > 1 else \
            features[0]

        q = self.out_layer(features_cat)
        return q


def num2tuple(num):
    return tuple(num) if isinstance(num, (tuple, list)) else (num, num)


import csv
import cv2
from pathlib import Path
import json

import torch
from tqdm import tqdm
import numpy as np
import math

import gym
from navsim.agent.ddpg import DDPGAgent
from navsim.util import sizeof_fmt, image_layout, s_hwc_to_chw
from navsim.util.dict import ObjDict
from navsim.env import env_info
from navsim.util import ResourceCounter

import traceback
from torch.utils.tensorboard import SummaryWriter


class TorchJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, torch.Tensor):
            return obj.tolist()
        # elif isinstance(obj, np.floating):
        #    return float(obj)
        # elif isinstance(obj, np.ndarray):
        #    return obj.tolist()
        else:
            return super(TorchJSONEncoder, self).default(obj)


"""
def s_hwc_to_chw(s):
    # TODO: HWC to CHW conversion optimized here
    # because pytorch can only deal with images in CHW format
    # we are making the optimization here to convert from HWC to CHW format.
    if isinstance(s, np.ndarray) and (s.ndim > 2):
        s = image_layout(s, 'hwc', 'chw')
    elif isinstance(s, list):  # state is a list of states
        for i in range(len(s)):
            if isinstance(s[i], np.ndarray) and (s[i].ndim > 2):
                s[i] = image_layout(s[i], 'hwc', 'chw')
    return s
"""


class Executor:
    """
        TODO
    """

    def __init__(self, resume=False, conf=None):
        """
        :param conf: A ObjDict containing dictionary and object interface
        :param env: Any gym compatible environment
        :param resume: True: means continue if run exists, else start new
                           False: means overwrite if exists, else start new
        """
        self.run_id = conf["run_config"]["run_id"]

        # self.conf = conf
        self.resume = resume

        self.run_base_folder = Path(self.run_id).resolve()
        self.run_base_folder_str = str(self.run_base_folder)
        #        if run_base_folder.is_dir():
        #            raise ValueError(f"{run_base_folder_str} exists as a non-directory. "
        #                             f"Please remove the file or use a different run_id")
        if resume and self.run_base_folder.is_dir():
            # self.conf = ObjDict().load_from_json_file(f"{self.run_base_folder_str}/conf.json")
            self.file_mode = 'a+'

        # else just start fresh
        else:

            self.run_base_folder.mkdir(parents=True, exist_ok=True)
            conf.save_to_json_file(str(self.run_base_folder / "conf.json"))
            self.file_mode = 'w+'

        self.run_config = ObjDict(conf.run_config)
        self.env_config = conf.get("env_config", None)
        if self.env_config is not None:
            self.env_config = ObjDict(self.env_config)
            env_log_folder = self.run_base_folder / 'env_log'
            self.env_config["log_folder"] = str(
                self.run_base_folder / 'env_log')
            if not resume:
                import shutil
                if env_log_folder.exists():
                    shutil.rmtree(env_log_folder)
                env_log_folder.mkdir(parents=True, exist_ok=True)

        pylog_filename = self.run_base_folder / 'py.log'  # TODO: use logger
        self.pylog_filename = str(pylog_filename)
        step_results_filename = self.run_base_folder / 'step_results.csv'
        self.step_results_filename = str(step_results_filename)
        episode_results_filename = self.run_base_folder / 'episode_results.csv'
        self.episode_results_filename = str(episode_results_filename)

        self.episode_num_filename = str(
            self.run_base_folder / "last_checkpointed_episode_num.txt")

        # TODO: Add the code to delete previous files
        # TODO: Add the code to add categories

        tb_folder = self.run_base_folder / 'tb'
        agent_folder = self.run_base_folder / 'agent'
        if not resume:
            import shutil
            for folder in [tb_folder, agent_folder]:
                if folder.exists():
                    shutil.rmtree(folder)
                folder.mkdir(parents=True, exist_ok=True)

        self.agent_folder = agent_folder

        self.summary_writer = SummaryWriter(f"{str(tb_folder)}")

        self.rc = ResourceCounter()
        self.files_open()

        try:
            if resume and self.run_base_folder.is_dir():
                with open(self.episode_num_filename,
                          mode='r') as episode_num_file:
                    episode_num = int(episode_num_file.read())
                    if self.env_config is not None:
                        self.env_config["start_from_episode"] = episode_num + 1
            else:
                if self.env_config is not None:
                    self.env_config["start_from_episode"] = 1
            self.env = None
            self.env_open()

            self.max_action = self.env.action_space.high[0]

            # Let us set the GPU or CPU

            if torch.cuda.is_available():
                torch.set_num_threads(1)
                torch.set_num_interop_threads(1)
                self.device = torch.device(
                    f"cuda:{self.run_config['agent_gpu_id']}")
                torch.cuda.set_device(self.run_config['agent_gpu_id'])
                torch.cuda.empty_cache()
            else:
                self.device = torch.device('cpu')
                torch.set_num_threads(8)
                torch.set_num_interop_threads(8)
            print(f'Torch is using {self.device}, '
                  f'{torch.get_num_threads()} threads and '
                  f'{torch.get_num_interop_threads()} interop threads')

            if self.run_config["mem_backend"] == "cupy":
                if torch.cuda.is_available():
                    from navsim.memory import CupyMemory
                    CupyMemory.set_device(self.run_config['agent_gpu_id'])
                    mem_backend = CupyMemory
                else:
                    raise ValueError(
                        "mem_backend=cupy but GPU is not available")
            elif self.run_config["mem_backend"] == "numpy":
                if torch.cuda.is_available():
                    print("Warning: GPU is available but mem_backend=numpy")
                from navsim.memory import NumpyMemory
                mem_backend = NumpyMemory

            obs_shapes = [obs.shape for obs in self.env.observation_space.spaces] if hasattr(self.env.observation_space,'spaces') else [self.env.observation_space.shape]

            if resume and self.run_base_folder.is_dir() and (
                    not self.run_config["clear_memory"]):
                memory_filename = str(
                    self.agent_folder / f"{episode_num}_memory.pkl")
                self.memory = mem_backend.load_from_pkl(memory_filename)
            else:
                memory_observation_space_shapes = []
                for o_shape in obs_shapes:
                    if len(o_shape) == 3:
                        # means its an image in HWC
                        # we pass the shapes as CWH
                        memory_observation_space_shapes.append(
                            (o_shape[2], o_shape[1], o_shape[0]))
                    else:
                        memory_observation_space_shapes.append(o_shape)
                self.memory = mem_backend(
                    capacity=self.run_config.memory_capacity,
                    state_shapes=memory_observation_space_shapes,
                    action_shape=self.env.action_space.shape,
                    seed=self.run_config['seed']
                )
            self.memory.info()

            self.agent = DDPGAgent(
                observation_space_shapes = obs_shapes,
                action_space_shape = self.env.action_space.shape,
                action_max=self.env.action_space.high[0],
                device=self.device,
                discount=self.run_config['discount'], tau=self.run_config['tau']
            )
            if resume and self.run_base_folder.is_dir():
                model_filename = str(
                    self.agent_folder / f"{episode_num}_model_state.pt")
                self.agent.load_checkpoint(model_filename)
            # TODO: self.agent.info()

            dummy_obs = self.agent.get_dummy_state()[0]
            dummy_act = self.agent.get_dummy_actions()[0]

            dummy_nn = self.agent.get_nn_wrapper()

            self.summary_writer.add_graph(dummy_nn,[dummy_obs, dummy_act])
            del dummy_nn, dummy_obs, dummy_act

            torch.manual_seed(self.run_config['seed'])
            np.random.seed(self.run_config['seed'])

        except Exception as e:
            self.env_close()
            self.files_close()
            print(traceback.format_exc())
            # print(e)

    # TODO: Find a better name for  this function
    def write_tb(self, group, values, t: int):
        for key, value in values.items():
            self.summary_writer.add_scalar(f'{group}/{key}', value, t)
            self.summary_writer.flush()

    def files_open(self):
        self.pylog_file = open(self.pylog_filename, mode=self.file_mode)

        self.step_results_file = open(self.step_results_filename,
                                      mode=self.file_mode)
        self.step_results_writer = csv.writer(self.step_results_file,
                                              delimiter=',',
                                              quotechar='"',
                                              quoting=csv.QUOTE_MINIMAL)

        self.episode_results_file = open(self.episode_results_filename,
                                         mode=self.file_mode)
        self.episode_results_writer = csv.writer(self.episode_results_file,
                                                 delimiter=',',
                                                 quotechar='"',
                                                 quoting=csv.QUOTE_MINIMAL)
        if not self.resume:
            self.step_results_writer.writerow(
                ['episode_num', 't', 'r', 'step_time', 'env_step_time',
                 'memory_append_time', 'memory_sample_time',
                 'agent_train_time', 'current_memory'])

            self.episode_results_writer.writerow(
                ['episode_num',
                 'episode_reward',
                 'episode_time',
                 'episode_peak_memory'])

        self.step_results_file.flush()
        self.episode_results_file.flush()

    def files_close(self):
        self.pylog_file.close()
        self.step_results_file.close()
        self.episode_results_file.close()

    def env_open(self):
        self.rc.start()

        if self.env_config is None:
            self.env = gym.make(self.run_config.env)
        else:
            self.env = gym.make(self.run_config.env, env_config=self.env_config)

        time_since_start, current_memory, peak_memory = self.rc.stop()
        log_str = f'env creation resource usage: \n' \
                  f'time:{time_since_start},' \
                  f'peak_memory:{sizeof_fmt(peak_memory)},' \
                  f'current_memory:{sizeof_fmt(current_memory)}\n'
        self.pylog_file.write(log_str)
        print(log_str)
        self.pylog_file.flush()
        #try:
        #    self.env.info()
        #except:
        env_info(self.env)

        # self.env.info_steps(save_visuals=True)

    def env_close(self):
        if self.env is None:
            print("Env is None")
        else:
            print("closing the env now")
            self.env.close()

    def execute(self):
        """
        Execute for the number of episodes
        :return:
        """
        start_from_episode = 1 if self.env_config is None else self.env_config[
            "start_from_episode"]
        t_max = int(self.run_config['episode_max_steps'])
        total_episodes = int(self.run_config['total_episodes'])
        num_episodes = total_episodes - (start_from_episode - 1)
        train_interval = int(self.run_config['train_interval'])
        checkpoint_interval = int(self.run_config['checkpoint_interval'])
        num_episode_blocks = int(math.ceil(num_episodes / checkpoint_interval))

        batch_size = self.run_config['batch_size']
        # save the state json at start of run
        model_filename = str(self.run_base_folder /
                             f"{start_from_episode}_" \
                             f"{total_episodes}_start_agent_state.json")
        json.dump(self.agent.get_state_dict(), open(model_filename, 'w'),
                  indent=2, sort_keys=True, cls=TorchJSONEncoder)

        #        print("Debugging training loop")
        #        print(f"{num_episode_blocks},{num_episodes},{self.conf.env_config['start_from_episode']}")
        t_global = 1

        episode_resources = np.full((checkpoint_interval, 2, 3), 0.0)
        # [[[0] * 3] * 2] * checkpoint_interval
        episode_steps = [0] * checkpoint_interval
        step_res = np.full((checkpoint_interval, t_max, 10, 3), 0.0)
        # [[[[0] * 3] * 10] * t_max] * checkpoint_interval
        step_rew = np.full((checkpoint_interval, t_max), 0.0)
        step_spl = np.full((checkpoint_interval, t_max), 0.0)
        step_loss = np.full((checkpoint_interval, t_max, 2), 0.0)

        for i in range(0, num_episode_blocks):
            start_episode = (start_from_episode - 1) + (
                    (i * checkpoint_interval) + 1)
            stop_episode = min((start_from_episode - 1)
                               + ((i + 1) * checkpoint_interval),
                               total_episodes)
            episodes_in_block = stop_episode - start_episode + 1
            # episode_rewards = np.full(episodes_in_block, 0.0)
            ckpt_e = 0
            ckpt_t_global = t_global

            for episode_num in tqdm(
                    iterable=range(start_episode, stop_episode + 1),
                    desc=f"Episode {start_episode}-{stop_episode}/{total_episodes}",
                    unit='episode', ascii=True, ncols=80,
                    bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} [{rate_fmt}{postfix}]'):
                self.rc.start()
                episode_counter = episode_num - start_episode
                episode_resources[ckpt_e, 0] = self.rc.snapshot()  # e0

                # initialise the episode counters
                episode_done = False
                # episode_return = 0
                t = 1

                # self.env.start_navigable_map()
                # observe initial s
                s = self.env.reset()  # s comes out of env as a tuple always
                # because pytorch can only deal with images in CHW format
                # we are making the optimization here to convert from HWC to CHW format.
                s = s_hwc_to_chw(s)

                # self.env.step(self.env.action_space.sample())
                # get the navigable map and save it as image
                # navigable_map = self.env.get_navigable_map()
                # if navigable_map is not None:
                #    cv2.imwrite(str((self.run_base_folder / f'navigable_map_{episode_num}.jpg').resolve()),navigable_map*255)
                # else:
                #    print(f'Map for episode {episode_num} is None')
                # step_res = [[0, 0, 0]] * 10

                while not episode_done:
                    # s0: start of overall_step
                    # s1,2: before and after memory.sample()
                    # s3,4: before and after agent.train()
                    # s5,6: before and after env.step()
                    # s7,8: before and after memory.append()
                    # s9: end of step
                    step_res[ckpt_e, t - 1, 0] = self.rc.snapshot()  # s0

                    # do the random sampling until enough memory is full
                    if self.memory.size < batch_size:
                        a = self.env.action_space.sample()
                        # rescale break between 0,1 from -1,1
                        # NewValue = (((OldValue - OldMin) * (NewMax - NewMin)) / (OldMax - OldMin)) + NewMin
                        # a[2] = (a[2] + 1) / 2
                        step_res[ckpt_e, t - 1, 1:5] = [[0.0] * 3] * 4
                        # s1-4
                    else:
                        # TODO: Find the best place to train, moved here for now
                        if train_interval and (
                                (t_global % train_interval) == 0):
                            step_res[ckpt_e, t - 1, 1] = self.rc.snapshot()
                            # s1
                            batch_s, batch_a, batch_r, batch_s_, batch_d = \
                                self.memory.sample(batch_size)
                            step_res[ckpt_e, t - 1, 2] = self.rc.snapshot()
                            # s2
                            # print('training the agent')
                            step_res[ckpt_e, t - 1, 3] = self.rc.snapshot()
                            # s3
                            self.agent.train(batch_s, batch_a, batch_r,
                                             batch_s_, batch_d)
                            step_res[ckpt_e, t - 1, 4] = self.rc.snapshot()
                            # s4
                        else:
                            step_res[ckpt_e, t - 1, 1:5] = [[0.0] * 3] * 4
                            # s1-4

                        a = (self.agent.select_action(s) + np.random.normal(
                            0, self.max_action * self.run_config['expl_noise'],
                            size=self.env.action_space.shape[0])
                             ).clip(
                            -self.max_action,
                            self.max_action
                        )

                    step_res[ckpt_e, t - 1, 5] = self.rc.snapshot()  # s5
                    s_, r, episode_done, info = self.env.step(a)
                    step_res[ckpt_e, t - 1, 6] = self.rc.snapshot()  # s6

                    # because pytorch can only deal with images in CHW format
                    # we are making the optimization here to convert from HWC to CHW format.
                    s_ = s_hwc_to_chw(s_)

                    #    s = [[s]]  # make state a list of sequences
                    #    s_ = [[s_]]
                    # elif isinstance(s[0], (int,float)):  # state does not have multiple seq
                    # s = [s]
                    # s_ = [s_]
                    # for item in s_:

                    step_res[ckpt_e, t - 1, 7] = self.rc.snapshot()
                    # s7
                    self.memory.append(
                        s=s, a=a, s_=s_, r=r,
                        d=float(episode_done))  # if t < t_max -1 else 1)
                    s = s_
                    step_res[ckpt_e, t - 1, 8] = self.rc.snapshot()  # s8

                    step_rew[ckpt_e, t - 1] = r
                    step_spl[ckpt_e, t - 1] = self.env.spl_current
                    step_loss[ckpt_e, t - 1] = [
                        0 if self.agent.actor_loss is None else self.agent.actor_loss.data.cpu().numpy(),
                        0 if self.agent.critic_loss is None else self.agent.critic_loss.data.cpu().numpy()]

                    # print(step_loss[ckpt_e, t - 1])
                    # episode_rewards[ckpt_e] += r

                    # if self.memory.size >= self.run_config['batch_size'] * self.run_config['batches_before_train']:

                    #                    if (t >= self.config['batch_size'] * self.config['batches_before_train']) and (t % 1000 == 0):
                    # episode_evaluations.append(evaluate_policy(self.agent, self.env, self.config['seed']))

                    step_res[ckpt_e, t - 1, 9] = self.rc.snapshot()  # s9
                    # print(ckpt_e,t-1, step_res[ckpt_e][t-1])
                    # print(t,step_res[0][0])
                    if episode_done or (t_max and (t >= t_max)):
                        break
                    else:
                        t += 1
                        t_global += 1
                # print(step_res[0][0])
                # end of while loop for one episode
                # episode end processing
                episode_resources[ckpt_e, 1] = self.rc.stop()  # e1
                episode_steps[ckpt_e] = t
                ckpt_e += 1

                #            if self.enable_logging:
                #                self.writer.add_scalar('Episode Reward', episode_return, t)
                #            episode_rewards.append(episode_return)

                # s = self.env.reset()[0]
                # episode_done = False
                # episode_return = 0
                # episode_timesteps = 0
                # episode_num += 1
            # end of tqdm-for loop - checkpoint block of episodes
            # now lets checkpoint everything

            t_global = ckpt_t_global
            # model and memory checkpoint
            with open(self.episode_num_filename, mode='w') as episode_num_file:
                episode_num_file.write(str(episode_num))
            model_filename = str(
                self.agent_folder / f"{episode_num}_model_state.pt")
            memory_filename = str(
                self.agent_folder / f"{episode_num}_memory.pkl")
            self.agent.save_checkpoint(model_filename)
            self.memory.save_to_pkl(memory_filename)

            # print(step_res[0][0])
            # episode data checkpoint
            for e_num in range(episodes_in_block):
                episode_time = episode_resources[e_num, 1, 0] - \
                               episode_resources[e_num, 0, 0]
                episode_return = step_rew[e_num, 0:episode_steps[e_num]].sum()
                self.write_tb('episode',
                              {'return': episode_return,
                               'time': episode_time,
                               'peak_memory': episode_resources[e_num, 1, 2],
                               'total_steps': episode_steps[e_num],
                               'max_spl': step_spl[e_num].max(),
                               'min_spl': step_spl[e_num].min(),
                               'spl_span': step_spl[e_num].max() - step_spl[
                                   e_num].min()
                               },
                              start_episode + e_num)  # episode_num
                self.episode_results_writer.writerow(
                    [start_episode + e_num,  # episode_num
                     episode_return,
                     episode_time,
                     episode_resources[e_num, 1, 2]])  # episode_peak_memory

                # step data checkpoint
                for t in range(episode_steps[e_num]):
                    # s0: start of overall_step
                    # s1,2: before and after memory.sample()
                    # s3,4: before and after agent.train()
                    # s5,6: before and after env.step()
                    # s7,8: before and after memory.append()
                    # s9: end of step
                    # print(e_num,t,step_res[e_num][t])
                    step_time = step_res[e_num, t, 9, 0] - step_res[
                        e_num, t, 0, 0]
                    memory_sample_time = step_res[e_num, t, 2, 0] - step_res[
                        e_num, t, 1, 0]
                    agent_train_time = step_res[e_num, t, 4, 0] - step_res[
                        e_num, t, 3, 0]
                    env_step_time = step_res[e_num, t, 6, 0] - step_res[
                        e_num, t, 5, 0]
                    memory_append_time = step_res[e_num, t, 8, 0] - step_res[
                        e_num, t, 7, 0]
                    # peak_memory = step_res[3][2]
                    current_memory = step_res[e_num, t, 9, 1]
                    r = step_rew[e_num, t]
                    self.write_tb('step',
                                  {'step_reward': r,
                                   'step_spl': step_spl[e_num, t]
                                   },
                                  t_global)
                    self.write_tb('time',
                                  {
                                      'step_time': step_time,
                                      'env_step_time': env_step_time,
                                      'memory_append_time': memory_append_time,
                                      'memory_sample_time': memory_sample_time,
                                      'agent_train_time': agent_train_time,
                                      'current_memory': current_memory,
                                  },
                                  t_global)
                    self.write_tb('loss',
                                  {
                                      'actor_loss': step_loss[e_num, t, 0],
                                      'critic_loss': step_loss[e_num, t, 1],
                                  },
                                  t_global)
                    self.step_results_writer.writerow(
                        [start_episode + e_num, t + 1, r, step_time,
                         env_step_time,
                         memory_append_time, memory_sample_time,
                         agent_train_time, current_memory])
                    t_global += 1
                # steps checkpoint loop finishes
            # episodes checkpoint loop finishes
            self.episode_results_file.flush()
            self.step_results_file.flush()
            ckpt_e = 0
        # save the state json at end of run
        model_filename = f"{self.run_base_folder_str}/" \
                         f"{start_from_episode}_" \
                         f"{total_episodes}_stop_agent_state.json"
        json.dump(self.agent.get_state_dict(), open(model_filename, 'w'),
                  indent=2, sort_keys=True, cls=TorchJSONEncoder)

        self.agent.save_to_onnx(folder=self.run_base_folder_str, critic=False)
        return



def env_info(env):
    """Prints the information about the environment

    """
    print('-----------')
    print("Env Info")
    print('-----------')
    if env.spec is not None:
        print(env.spec.id)
    print('Action Space <gym_type(low,high,shape,type)>:', env.action_space)
    # TODO: print env_config if it exists
    # print('Observation Mode:', env.obs_mode)
    # print('Gym Observation Space:', self.genv.observation_space)
    # print('Gym Observation Space Shape:', self.genv.observation_space.shape)
    print('Observation Space:', env.observation_space)
    if hasattr(env.observation_space, 'spaces'):
        print('Observation Space Spaces:',
              [obs for obs in env.observation_space.spaces])
    #    print('Observation Space Types:', [type(obs) for obs in env.observation_space.spaces])
    print('Reward Range:', env.reward_range)
    print('Metadata:', env.metadata)
    print('--------------------------------------')


import torch
import torch.nn.functional as F

from copy import deepcopy

# for actor:
#   n_x is state dimension
#   n_y is action dimension
#   output: max_action * tanh(fc3)
#   returns a
# for critic:
#   n_x is state_dimension + action_dim
#   n_y is 1
#   input:  cat([state,action],1)
#   returns q


from navsim.agent.nn import ActorCriticWrapper, Actor, Critic


class DDPGAgent(object):
    NN_WRAPPER = ActorCriticWrapper

    def __init__(self, observation_space_shapes, action_space_shape, action_max,
                 device, discount=0.99, tau=0.005):
        # self.env = env
        self.state_dims = observation_space_shapes
        self.action_dim = action_space_shape[0]
        self.action_max = action_max
        self.device = device
        self.discount = discount
        self.tau = tau
        # self.state_dim = state_dim

        # self.max_action = self.env.action_space.high[0]

        # if self.env.obs_mode == 0:
        #    self.vector_state_dimension = self.env.state_dims[0][0]
        # elif self.env.obs_mode == 1:
        #    self.vector_state_dimension = None
        # elif self.env.obs_mode == 2:
        #    self.vector_state_dimension = self.env.state_dims[3][0]

        self.actor = Actor(self.state_dims,
                           self.action_dim,
                           self.action_max).to(device)
        self.critic = Critic(self.state_dims,
                             self.action_dim).to(device)

        self.actor_target = deepcopy(self.actor)
        self.actor_optimizer = torch.optim.Adam(self.actor.parameters())

        self.critic_target = deepcopy(self.critic)
        self.critic_optimizer = torch.optim.Adam(self.critic.parameters())

        self.actor_loss = None
        self.critic_loss = None

    def select_action(self, state):
        # if self.env.obs_mode == 0:
        #    state = torch.FloatTensor(state[0].reshape(1, -1)).to(self.device)
        state = [torch.as_tensor(s,
                                 dtype=torch.float,
                                 device=self.device).unsqueeze(0) for s in
                 state]
        return self.actor(state).data.cpu().numpy().flatten()

    @staticmethod
    def soft_update(local_model, target_model, tau):
        for target_param, local_param in zip(target_model.parameters(),
                                             local_model.parameters()):
            target_param.data.copy_(
                tau * local_param.data + (1.0 - tau) * target_param.data)

    def get_state_dict(self):
        state = {
            "critic": self.critic.state_dict(),
            "critic_optimizer": self.critic_optimizer.state_dict(),
            "actor": self.actor.state_dict(),
            "actor_optimizer": self.actor_optimizer.state_dict()
        }
        return state

    def save_checkpoint(self, filename):

        torch.save(self.get_state_dict(), filename)
        """
        torch.save(self.critic.state_dict(), filename + '_critic')
        torch.save(self.critic_optimizer.state_dict(), filename + '_critic_optimizer')
        torch.save(self.actor.state_dict(), filename + '_actor')
        torch.save(self.actor_optimizer.state_dict(), filename + '_actor_optimizer')
        """

    def save_to_onnx(self, folder='.', critic=False):
        #        device = next(model.parameters()).device
        device = self.device

        # prepare input data
        input_data = []
        input_names = []
        for state_dim in self.state_dims:
            if len(state_dim) == 1:
                random_input = torch.randn(1, state_dim[0]).to(device)
                input_name = f'state_{state_dim[0]}'
            else:  # visual
                random_input = torch.randn(1, state_dim[2], state_dim[0],
                                           state_dim[1]).to(device)
                input_name = f'state_{state_dim[0]}_{state_dim[1]}_{state_dim[2]}'

            input_data.append(random_input)
            input_names.append(input_name)

        # export actor
        model = self.actor
        torch.onnx.export(model,
                          args=input_data,
                          f=f"{folder}/actor.onnx",
                          export_params=True,
                          opset_version=9,
                          do_constant_folding=True,
                          input_names=input_names,
                          output_names=['action'])

        if critic:
            # add action data for critic
            action_dim = self.action_dim[0]
            random_input = torch.randn(1, action_dim).to(device)
            input_name = f'action_{action_dim}'
            input_data = [input_data]

            # print(len(input_data))
            input_data.append(random_input)
            input_data = tuple(input_data)
            # print(len(input_data))
            input_names.append(input_name)

            # export critic
            model = self.critic
            torch.onnx.export(model,
                              args=input_data,
                              f=f"{folder}/critic.onnx",
                              export_params=True,
                              opset_version=9,
                              do_constant_folding=True,
                              input_names=input_names,
                              output_names=['q'])

    def load_checkpoint(self, filename):
        state = torch.load(filename)
        self.critic.load_state_dict(state['critic'])
        self.critic_optimizer.load_state_dict(state['critic_optimizer'])
        self.actor.load_state_dict(state['actor'])
        self.actor_optimizer.load_state_dict(state['actor_optimizer'])

        """
        self.critic.load_state_dict(
            torch.load(
                filename + "_critic",
                map_location=torch.device('cpu')
            )
        )
        self.critic_optimizer.load_state_dict(
            torch.load(
                filename + "_critic_optimizer",
                map_location=torch.device('cpu')
            )
        )
        self.critic_target = deepcopy(self.critic)
        self.actor.load_state_dict(
            torch.load(
                filename + "_actor",
                map_location=torch.device('cpu')
            )
        )
        self.actor_optimizer.load_state_dict(
            torch.load(
                filename + "_actor_optimizer",
                map_location=torch.device('cpu')
            )
        )
        """
        self.actor_target = deepcopy(self.actor)

    def train(self, state, action, reward, next_state, episode_done):
        # print('agent train state ',state.shape)
        # print('agent train next_state',next_state.shape)
        # convert state to list of tensors on GPU
        state = [torch.as_tensor(s,
                                 dtype=torch.float,
                                 device=self.device) for s in state]
        next_state = [torch.as_tensor(s,
                                      dtype=torch.float,
                                      device=self.device) for s in next_state]
        action = torch.as_tensor(action, dtype=torch.float, device=self.device)
        target_q = self.critic_target(next_state, self.actor_target(next_state))

        reward = torch.as_tensor(reward, dtype=torch.float, device=self.device)
        episode_done = torch.as_tensor(episode_done, dtype=torch.float,
                                       device=self.device)
        # reward = target_q.new(reward)
        # episode_done = target_q.new(episode_done)
        target_q = reward + (
                (1.0 - episode_done) * self.discount * target_q).detach()

        current_q = self.critic(state, action)
        self.critic_loss = F.mse_loss(current_q, target_q)
        self.critic_optimizer.zero_grad()
        self.critic_loss.backward()
        self.critic_optimizer.step()
        self.actor_loss = -self.critic(state, self.actor(state)).mean()
        self.actor_optimizer.zero_grad()
        self.actor_loss.backward()
        self.actor_optimizer.step()
        DDPGAgent.soft_update(self.critic, self.critic_target, self.tau)
        DDPGAgent.soft_update(self.actor, self.actor_target, self.tau)

    def info(self):
        print("Agent Info")
        print('-----------')
        print('Not implemented yet')

    def get_nn_wrapper(self):
        return ActorCriticWrapper(self.state_dims,
                                  self.action_dim,
                                  self.action_max).to(self.device)

    def get_dummy_state(self):
        # prepare input data
        state_names = []
        state_data = []
        for state_dim in self.state_dims:
            # obs_dim = [1] + [state_dim[2],state_dim[1],state_dim[0],
            #hwc to chw
            state_d = (state_dim[2],state_dim[1],state_dim[0]) if len(state_dim)==3 else state_dim
            obs = torch.zeros([1] + list(state_d), dtype=torch.float,
                              device=self.device)
            obs_name = 'state'
            for i in range(len(state_d)):
                #    input_dim.append(state_dim[i])
                obs_name += f'_{state_d[i]}'
            state_names.append(obs_name)
            state_data.append(obs)

        # print([o.shape for o in state_data],state_names)
        return state_data, state_names

    def get_dummy_actions(self):
        actions_data = torch.zeros([1, self.action_dim], dtype=torch.float, device=self.device)
        actions_names = f'action_{self.action_dim}'
        return actions_data, actions_names


def evaluate_policy(policy, env, seed, eval_episodes=10, render=False):
    eval_env = env
    eval_env.seed(seed + 100)
    avg_reward = 0.
    for _ in range(eval_episodes):
        state, done = eval_env.reset(), False
        while not done:
            action = policy.select_action(state)
            if render:
                eval_env.render()
            state, reward, done, _ = eval_env.step(action)
            avg_reward += reward
    avg_reward /= eval_episodes
    return avg_reward



import unittest

import torch


class TestTorch(unittest.TestCase):

    def test_torch_cuda_available(self):
        self.assertTrue(torch.cuda.is_available())


if __name__ == '__main__':
    unittest.main()



import pickle
from abc import ABC
from typing import Union, Optional

import cupy as cp
import numpy as np


class Memory(ABC):
    def __init__(self, capacity: int,
                 state_shapes: Union[list, tuple, int],
                 action_shape: Union[list, tuple, int],
                 seed: Optional[float] = None,
                 backend: object = np):
        self.capacity = capacity
        self.ptr: int = 0
        self.size: int = 0
        self.seed: Optional[float] = seed
        self.backend = backend
        self.a = self.r = self.d = None

        # if isinstance(state_shapes, (list, tuple)):  # state_dims is an int
        if isinstance(state_shapes, (int, float)):  # state_dims is an int
            state_shapes = [[state_shapes]]  # make it a list of sequences
        elif isinstance(state_shapes[0], (int, float)):
            # state_dims 1st member is an int
            state_shapes = [state_shapes]  # make it a list of sequences
        # else state_dims is a list of sequences,
        # i.e. [[state_shape], [state_shape], [state_shape]]

        self.s_len: int = len(state_shapes)
        self.s = [None] * self.s_len
        self.s_ = [None] * self.s_len

        self.rng = self.backend.random.default_rng(seed=seed)

        for i in range(self.s_len):
            s_shapes = [capacity] + list(state_shapes[i])
            self.s[i] = self.backend.full(s_shapes, 0.0)
            self.s_[i] = self.backend.full(s_shapes, 0.0)

        a_shape = [capacity] + list(action_shape)
        self.a = self.backend.full(a_shape, 0.0)

        self.r = self.backend.full((capacity, 1), 0.0)
        self.d = self.backend.full((capacity, 1), False)

    def __str__(self):
        return f'module: {type(self).__name__} \n' \
               f'capacity: {self.capacity} \n' \
               f'size: {self.size} \n' \
               f'current pointer: {self.ptr} \n' \
               f'seed: {self.seed} \n' \
               f'Shapes and types: \n' \
               f's: {[f"{i.shape} of {type(i)}" for i in self.s]} \n' \
               f'a: {self.a.shape} {type(self.a)} \n' \
               f'r: {self.r.shape} {type(self.r)} \n' \
               f's_: {[f"{i.shape} of {type(i)}" for i in self.s_]} \n' \
               f'd: {self.d.shape} {type(self.d)} \n'

    def info(self):
        print('-----------')
        print("Memory Info")
        print('-----------')
        print(self)
        print(f'----------------------------------------')

    @staticmethod
    def sample_info(s, a, r, s_, d):
        print('-----------')
        print("Sample Info")
        print('-----------')
        print("Shapes and types:")
        print('s:', [f"{i.shape} of {type(i)}" for i in s])
        print('a:', a.shape, type(a))
        print('r:', r.shape, type(r))
        print('s_:', [f"{i.shape} of {type(i)}" for i in s_])
        print('d:', d.shape, type(d))
        print('-----------')

    def sample(self, size):
        if (size == 0) or (size >= self.size):
            # idx = list(range(self.size))   # just return all
            return self.s, self.a, self.r, self.s_, self.d
        else:
            idx = self.rng.integers(low=0, high=self.size, size=size)
            return [self.s[i][idx] for i in range(self.s_len)], \
                   self.a[idx], \
                   self.r[idx], \
                   [self.s_[i][idx] for i in range(self.s_len)], \
                   self.d[idx]

    # def __len__(self):
    #    return len(self.mem)

    # def __repr__(self):
    #    return str(self.mem)

    # def __getitem__(self, sliced):
    #    return CupyMemory(capacity=self.capacity, seed=self.seed,
    #                      mem=self.mem[sliced])

    def save_to_pkl(self, filename):
        rng = self.rng
        self.rng = None
        backend = self.backend
        self.backend = None
        with open(filename, "wb") as out_file:
            pickle.dump(self, out_file)
        self.backend = backend
        self.rng = rng

    @staticmethod
    def load_from_pkl(filename):
        with open(filename, "rb") as in_file:
            return pickle.load(in_file)

    def append(self, s: Union[list, tuple, int, float],
               a: Union[list, tuple, int, float],
               r: float,
               s_: Union[list, tuple, int, float],
               d: float):
        if isinstance(s, (int, float)):  # state is an int
            s = [[s]]  # make state a list of sequences
            s_ = [[s_]]
        elif isinstance(s[0], (int, float)):  # state does not have multiple seq
            s = [s]
            s_ = [s_]
        for i in range(self.s_len):
            self.s[i][self.ptr] = self.backend.asarray(s[i], dtype=np.float32)
            self.s_[i][self.ptr] = self.backend.asarray(s_[i], dtype=np.float32)

        self.a[self.ptr] = self.backend.asarray(a, dtype=np.float32)
        self.r[self.ptr] = r
        self.d[self.ptr] = d

        self.ptr = (self.ptr + 1) % self.capacity
        self.size = min(self.size + 1, self.capacity)


class CupyMemory(Memory):

    def __init__(self, capacity: int,
                 state_shapes: Union[list, tuple, int],
                 action_shape: Union[list, tuple, int],
                 seed: Optional[float] = None):
        """

        :param capacity: total capacity of the memory
        :param state_shapes: a tuple of state dimensions
        :param action_shape: int representing number of actions
        :param seed: seed to provide to numpy
        """

        super().__init__(capacity, state_shapes, action_shape, seed, backend=cp)

    def __str__(self):
        return f'module: {type(self).__name__} \n' \
               f'cupy device id: {cp.cuda.Device().id} \n' \
               f'capacity: {self.capacity} \n' \
               f'size: {self.size} \n' \
               f'current pointer: {self.ptr} \n' \
               f'seed: {self.seed} \n' \
               f'Shapes and types: \n' \
               f's: {[f"{i.shape} {type(i)} {i.device}" for i in self.s]} \n' \
               f'a: {self.a.shape} {type(self.a)} {self.a.device} \n' \
               f'r: {self.r.shape} {type(self.r)} {self.r.device} \n' \
               f's_: {[f"{i.shape} {type(i)} {i.device}" for i in self.s_]} \n' \
               f'd: {self.d.shape} {type(self.d)} {self.d.device} \n'

    @staticmethod
    def set_device(self, gpu_id=0):
        cp.cuda.Device(gpu_id).use()

    @staticmethod
    def get_device(self):
        return cp.cuda.Device().id

    @staticmethod
    def load_from_pkl(filename):
        with open(filename, "rb") as in_file:
            mem = pickle.load(in_file)
        mem.backend = cp
        mem.rng = mem.backend.random.default_rng(seed=mem.seed)
        return mem


class NumpyMemory(Memory):
    """Stores and returns list of numpy arrays for s a r \s_ d

    assumes you are storing s a r \s_ d : state action reward next_state done
    # state / next_state : tuple of n-dim numpy arrays
    # a : 1-d numpy array of floats or ints
    # r : 1-d numpy array of floats
    # d : 1-d numpy array of booleans

    """

    def __init__(self, capacity: int,
                 state_shapes: Union[list, tuple, int],
                 action_shape: Union[list, tuple, int],
                 seed: Optional[float] = None):
        """

        :param capacity: total capacity of the memory
        :param state_shapes: a tuple of state dimensions
        :param action_shape: int representing number of actions
        :param seed: seed to provide to numpy
        """

        super().__init__(capacity, state_shapes, action_shape, seed)

        # from numpy.random import MT19937, RandomState, Generator
        # mt = MT19937(seed)
        # rs = RandomState(mt)
        # rg = Generator(mt)

        @staticmethod
        def load_from_pkl(filename):
            with open(filename, "rb") as in_file:
                mem = pickle.load(in_file)
            mem.backend = np
            mem.rng = mem.backend.random.default_rng(seed=mem.seed)
            return mem



import time
import argparse
from pathlib import Path
import subprocess
import os


def monitorsubprocesses(proc_list, timeout):
    while len(proc_list) != 0:
        for p in proc_list:
            status = p.poll()
            if status == 0:
                print("Success", p, len(proc_list))
                proc_list.remove(p)

            elif status != None:
                print("Failure", p, len(proc_list), status)
                proc_list.remove(p)
        time.sleep(timeout)


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('binary', metavar='B', type=Path,
                        default=None, #"/mnt/AICOOP_binaries/Build2.4.4/Berlin_Walk_V2.x86_64",
                        help="File path to Unity game binary")
    # TODO find this dynamically with nvidia-smi and parse
    # nvidia-smi  --help-query-gpu
    # p=subprocess.Popen(["nvidia-smi", '--query-gpu=count', '--format=csv', '-i', 0])
    # p=subprocess.Popen(["nvidia-smi", '--query-gpu=memory.total', '--format=csv', '-i', 0])
    parser.add_argument('-agent_x-disp', metavar='X', default="0", help="X DISPLAY port")
    parser.add_argument('--num-gpus', default=4, type=int, help="Number of available gpus")
    parser.add_argument('--max-proc-vram', default=5000, type=int, help="Maximum VRAM used by Unity Game")
    parser.add_argument('--single-gpu-vram', default=11178, type=int, help="Total amount of available vram per gpu")
    parser.add_argument('--timeout', default=30, type=int,
                        help="Amount of time in seconds between process status checks")
    parser.add_argument('--override-num-proc', type=int, help="Manually set number of processes per gpu")

    args = parser.parse_args()

    print(args)
    # "python ../benchmarks/benchmark.py  ../../AICOOP_binaries/Build2.4.4/Berlin_Walk_V2.x86_64 -a VectorVisual"
    unity_binary_path = args.binary

    gpu_env = os.environ.copy()

    # Start benchmark.py on each gpu as process
    proc_list = []
    proc_per_gpu = int(args.single_gpu_vram / args.max_proc_vram)
    if args.override_num_proc is not None:
        proc_per_gpu = args.override_num_proc

    for gpu_idx in range(0, args.num_gpus):
        for proc_idx in range(0, proc_per_gpu):
            # proc_list.append(subprocess.Popen(["ls", "-la"]))
            gpu_env["DISPLAY"] = ":" + str(args.x_disp) + "." + str(gpu_idx)
            sp_args = ["navsim-benchmark", unity_binary_path, "-a", "VectorVisual", "--worker_id",
                       str(proc_idx + (gpu_idx * proc_per_gpu))]
            print(sp_args)
            proc_list.append(subprocess.Popen(sp_args, env=gpu_env))

    # Monitor subprocess list by checking return codes on a loop
    monitorsubprocesses(proc_list, args.timeout)

    print("DONE")

# For python debugger to directly run this script
if __name__ == "__main__":
    main()




import numpy as np


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def image_layout(x, old: str, new: str):
    new = [old.index(char) for char in new]
    return x.transpose(new)


def hwc_to_chw(s):
    if s.ndim == 3:
        s_ = image_layout(s, 'hwc', 'chw')
    elif s.ndim == 4:
        s_ = image_layout(s, 'nhwc', 'nchw')
    return s_


def s_hwc_to_chw(s):
    # TODO: HWC to CHW conversion optimized here
    # because pytorch can only deal with images in CHW format
    # we are making the optimization here to convert from HWC to CHW format.
    if isinstance(s, np.ndarray) and (s.ndim > 2):
        s = hwc_to_chw(s)
    elif isinstance(s, list):  # state is a list of states
        for i in range(len(s)):
            if isinstance(s[i], np.ndarray) and (s[i].ndim > 2):
                s[i] = hwc_to_chw(s[i])
    return s


def env_state_shapes(env):
    return [obs.shape for obs in env.observation_space.spaces] \
        if hasattr(env,'spaces') else [env.observation_space.shape]



import argparse
from typing import Optional, List, Set

non_default_args: Set[str] = set()


class ArgAction(argparse.Action):
    """
    Class to handle actions.
    - Add to the non-default list.
    """

    def __call__(self, arg_parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)
        non_default_args.add(self.dest)


class ArgActionStoreTrue(ArgAction):
    """
    Class to handle actions for flags that need to be set to true
    """

    def __init__(self, nargs=0, **kwargs):
        super().__init__(nargs=nargs, **kwargs)

    def __call__(self, arg_parser, namespace, values, option_string=None):
        super().__call__(arg_parser, namespace, True, option_string)


def _create_argparser() -> argparse.ArgumentParser:
    argparser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    # argparser.add_argument(
    #    "config_file", nargs="?", default=None
    # )

    run_desc = "The arguments are used to configure the runtime execution."
    run_conf = argparser.add_argument_group(title="Run Configuration",
                                            description=run_desc)

    run_conf.add_argument(
        "--run_id",
        default="demo",
        help="The identifier for the training run. This identifier is used to name the "
             "subdirectories in which the trained model and summary statistics are saved as well "
             "as the saved model itself. If you use TensorBoard to view the training statistics, "
             "always set a unique run-id for each training run. (The statistics for all runs with the "
             "same id are combined as if they were produced by a the same session.)",
        action=ArgAction,
    )

    run_conf.add_argument(
        "--rl_backend",
        default=None,
        help="The backend library for RL.",
        action=ArgAction,
    )

    run_conf.add_argument(
        "--debug",
        default=False,
        help="Turn debugging on",
        action=ArgActionStoreTrue,
    )

    run_conf.add_argument(
        "--force",
        default=False,
        action=ArgActionStoreTrue,
        help="Whether to force-overwrite this run-id's existing summary and model data. (Without "
             "this flag, attempting to train a model with a run-id that has been used before will throw "
             "an error.",
    )

    run_conf.add_argument(
        "--total_episodes",
        default=2,
        action=ArgAction,
        help="Total number of episodes to run. If resume is used, then it will try to read the previously run episodes and continue from there.",
    )
    run_conf.add_argument(
        "--train_interval",
        default=16,
        help="Train the model after these many global steps. If set to 0 then model wont train",
        action=ArgAction,
    )

    run_conf.add_argument(
        "--checkpoint_interval",
        default=1,
        action=ArgAction,
        help="Execute the episodes in blocks of checkpoint intervals",
    )

    run_conf.add_argument(
        "--agent_gpu_id",
        default=0,
        action=ArgAction,
        help="Which GPU to run models in agent on",
    )

    run_conf.add_argument(
        "--episode_max_steps",
        default=100,
        action=ArgAction,
        help="Maximum number of steps in an Episode, aka Episode Length",
    )

    run_conf.add_argument(
        "--seed",
        default=0,
        action=ArgAction,
        help="Seed",
    )

    run_conf.add_argument(
        "--discount",
        default=0.99,
        action=ArgAction,
        help="discount",
    )
    run_conf.add_argument(
        "--tau",
        default=5e-3,
        action=ArgAction,
        help="tau",
    )
    run_conf.add_argument(
        "--exploration_noise",
        default=0.1,
        action=ArgAction,
        help="exploration_noise",
    )
    run_conf.add_argument(
        "--batch_size",
        default=32,
        action=ArgAction,
        help="Batch Size",
    )

    run_conf.add_argument(
        "--memory_capacity",
        default=100,
        action=ArgAction,
        help="Total capacity of memory, should be > batch_size",
    )

    run_conf.add_argument(
        "--resume",
        default=False,
        action=ArgActionStoreTrue,
        help="resumes the training from last saved checkpoint"
    )

    run_conf.add_argument(
        "--clear_memory",
        default=False,
        action=ArgActionStoreTrue,
        help="On resuming, the memory will be cleared and actions would be "
             "sampled from the neural network, if neural network is not trained "
             "yet, then actions would sample from action_space",
    )
    env_desc = "The arguments are used to configure the environment."
    env_conf = argparser.add_argument_group(title="Environment Configuration",
                                            description=env_desc)
    env_conf.add_argument(
        "--env",
        default=None,
        dest="env_path",
        help="Path to the Unity executable to train",
        action=ArgAction,
    )

    env_conf.add_argument(
        "--env_gpu_id",
        default=0,
        action=ArgAction,
        help="Which GPU to run env on",
    )
    env_conf.add_argument(
        "--timeout",
        default=600,
        help="TimeOut for the Env",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--obs_mode",
        default=0,
        help="Observation Mode : 0,1,2",
        action=ArgAction,
    )

    env_conf.add_argument(
        "--obs_height",
        default=256,
        help="Observation height",
        action=ArgAction,
    )

    env_conf.add_argument(
        "--obs_width",
        default=256,
        help="Observation width",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--segmentation_mode",
        default=1,
        help="Segmentation Mode : 1",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--task",
        default=0,
        help="Task",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--goal",
        default=0,
        help="Goal",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--goal_distance",
        default=10,
        help="Distance to goal from current location",
        action=ArgAction,
    )

    env_conf.add_argument(
        "--traffic_vehicles",
        default=0,
        help="Number of traffic vehicles",
        action=ArgAction,
    )

    env_conf.add_argument(
        "--agent_car_physics",
        default=0,
        help="Agent Car Physics Levels : 0,1,2,10",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--reward_for_goal",
        default=50,
        help="Reward for reaching the goal",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--reward_for_no_viable_path",
        default=-50,
        help="Reward for no more viable paths remaining to reach the goal",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--reward_step_mul",
        default=0.1,
        help="Multiply step reward [which is -(goal_reward / spl_start) with this number",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--reward_collision_mul",
        default=4,
        help="Multiply step reward with this number when there is a collision and add to reward",
        action=ArgAction,
    )

    env_conf.add_argument(
        "--reward_spl_delta_mul",
        default=1,
        help="Shortest path length delta multiplier",
        action=ArgAction,
    )
    env_conf.add_argument(
        "--save_actions",
        default=False,
        action=ArgActionStoreTrue,
        help="Save the actions vector at every step",
    )
    env_conf.add_argument(
        "--save_visual_obs",
        default=False,
        action=ArgActionStoreTrue,
        help="Save the visual observations at every step",
    )
    env_conf.add_argument(
        "--save_vector_obs",
        default=False,
        action=ArgActionStoreTrue,
        help="Save the vector observations at every step",
    )
    env_conf.add_argument(
        "--show_visual",
        default=False,
        action=ArgActionStoreTrue,
        help="Show visual obs window",
    )


    dev_desc = "The arguments are used by developers to benchmark and debug navsim API"
    dev_conf = argparser.add_argument_group(title="Developer Configuration",
                                            description=dev_desc)

    dev_conf.add_argument(
        "--mem_backend",
        default="cupy",
        help="The backend library for Rollback Memory.",
        action=ArgAction,
    )

    return argparser


argparser = _create_argparser()



import os

# Path(__file__).parent.joinpath('version.txt')
with open(os.path.join(os.path.dirname(__file__), 'version.txt'), 'r') as vf:
    __version__ = vf.read().strip()



from navsim import util, agent, executor, env


__all__ = ['agent', 'util', 'executor','env']



import setuptools

with open('navsim/version.txt', 'r') as vf:
    __version__ = vf.read().strip()
# from navsim.version import __version__

# TODO: Replace the long description comments once cleared
# with open("README.md", "r", encoding="utf-8") as fh:
#    long_description = fh.read()
long_description = " "

setuptools.setup(
    name="navsim",
    version=__version__,
    author="STTC, IST, UCF",
    author_email="armando@ucf.edu",
    description="Navigation Simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
    ],
    include_package_data=True,
    # data_files=[('navsim',['navsim/version.txt'])],
    python_requires='>=3.7',
    install_requires=[
        'numpy',
        # 'opencv-python',
        'gym>=0.18.0',
        'mlagents==0.27.0',
        'mlagents_envs==0.27.0',
        'gym_unity==0.27.0'
    ],
    entry_points={
        "console_scripts": [
            "navsim=navsim.main:main",
            "navsim-benchmark=navsim.benchmarks.benchmark:main",
            "navsim-saturate-gpu=navsim.benchmarks.saturate_gpu:main",
        ]
    },
)



import time
import tracemalloc
import gc

class ResourceCounter:
    """
    This class allows to monitor the following resources for the purpose of logging:
    1. Time:
    2. NumpyMemory

    Usage:
    1. Create the rc object:
    rc = ResourceCounter(clean=True)

    clean=True by default, does the garbage collection when you start the counter

    2. start the counter:
    rc.start()

    3. Take snapshot of current usage anytime:
    time_since_start, current_memory, peak_memory = rc.snapshot()

    4. Stop and get final values:
    time_since_start, current_memory, peak_memory = rc.snapshot()

    Can be restarted anytime from zero by calling start.

    """
    def __init__(self, clean:bool=False):
        # in fractional seconds
        self.start_time = 0
        self.stop_time = 0
        self.clean = clean

    def _check_and_raise_start(self):
        if self.start_time <= 0:
            raise ValueError('ResourceCounter in invalid state. Did you forget to start the counter?')

    def _check_and_raise_stop(self):
        if self.stop_time < self.start_time:
            raise ValueError('ResourceCounter in invalid state. Did you forget to stop the counter?')

    def start(self):
        if self.clean:
            gc.collect()
        self.start_time = time.process_time()
        self.stop_time = 0
        tracemalloc.start()

    def snapshot(self):
        self._check_and_raise_start()
        snapshot_time = time.process_time()
        time_since_start = snapshot_time - self.start_time
        current_memory, peak_memory = tracemalloc.get_traced_memory()
        return time_since_start, current_memory, peak_memory   # in seconds

    def stop(self):
        self._check_and_raise_start()
        time_since_start, current_memory, peak_memory = self.snapshot()
        self.stop_time = time_since_start + self.start_time
        tracemalloc.stop()
        if self.clean:
            gc.collect()
        return time_since_start, current_memory, peak_memory   # in seconds

    @property
    def elapsed_time(self):
        self._check_and_raise_start()
        self._check_and_raise_stop()

        return self.stop_time - self.start_time



from pathlib import Path
from typing import Optional, List, Set

import navsim
from .util.dict import ObjDict
from .cli_utils import argparser, non_default_args
from navsim.executor.navsim_executor import Executor


def main():
    """
    TODO: Implement configuration checks
    :return:
    """
    print(f'=========================================')
    print(f'Navsim Python API Version {navsim.__version__}')
    print(f'========================================')
    args = ObjDict(vars(argparser.parse_args()))
    print('arguments passed:')
    print(non_default_args)
    print("Passed + default arguments:")
    print(args.to_yaml())

    # lets get the arguments
    run_base_folder = Path(args.run_id).resolve()

    env_conf = ObjDict({
        "log_folder": str(run_base_folder / "env_log"),
        "env_path": args["env_path"],
        "worker_id": 0,
        "base_port": 5005,
        "seed": int(args["seed"]),
        "timeout": int(args["timeout"]),
        "obs_mode": int(args["obs_mode"]),
        "obs_height": int(args["obs_height"]),
        "obs_width": int(args["obs_width"]),
        "segmentation_mode": int(args["segmentation_mode"]),
        "episode_max_steps": int(args["episode_max_steps"]),
        "task": int(args["task"]),
        "goal": int(args["goal"]),
        "goal_distance": int(args["goal_distance"]),
        "traffic_vehicles": int(args["traffic_vehicles"]),
        "agent_car_physics": int(args["agent_car_physics"]),
        "reward_for_goal": float(args["reward_for_goal"]),
        "reward_for_no_viable_path": float(args["reward_for_no_viable_path"]),
        "reward_step_mul": float(args["reward_step_mul"]),
        "reward_collision_mul": float(args["reward_collision_mul"]),
        "reward_spl_delta_mul": float(args["reward_spl_delta_mul"]),
        "env_gpu_id": int(args["env_gpu_id"]),
        "debug": args["debug"],
        "save_actions": args["save_actions"],
        "save_vector_obs": args["save_vector_obs"],
        "save_visual_obs": args["save_visual_obs"],
        "show_visual": args["show_visual"]
    })

    run_conf = ObjDict({
        "run_id": args["run_id"],
        "env": 'navsim-v0',
        "agent_gpu_id": int(args["agent_gpu_id"]),
        "num_workers": 1,
        "episode_max_steps": int(args["episode_max_steps"]),
        "total_episodes": int(args["total_episodes"]),
        "seed": int(args["seed"]),
        "discount": float(args["discount"]),
        "tau": float(args["tau"]),
        "expl_noise": float(args["exploration_noise"]),
        "memory_capacity": int(args["memory_capacity"]),
        "batch_size": int(args["batch_size"]),
        "checkpoint_interval": int(args["checkpoint_interval"]),
        "train_interval": int(args["train_interval"]),
        "mem_backend": args["mem_backend"],
        "clear_memory": args["clear_memory"],
        "log_level": "DEBUG" if args.debug else "INFO",
        "framework": "torch"
    })

    if args.resume:
        if run_base_folder.exists():
            # TODO: Logic for resume here
            pass

        else:  # else just start fresh
            args.resume = False
            print(
                "Resume set to True, but nothing to resume from, starting fresh")

    # start fresh logic
    if not args.resume:
        run_base_folder.mkdir(parents=True, exist_ok=True)



    if args["rl_backend"] == "rllib":
        import ray.rllib.agents.ppo as ppo
        conf = ppo.DEFAULT_CONFIG.copy()
        # copy from run_config to conf
        for key in run_conf:
            if key in conf:
                conf[key] = run_conf[key]
        conf["ignore_worker_failures"] = True
    else:
        conf = ObjDict()
        conf['run_config'] = run_conf
    conf['env_config'] = env_conf

    # print("Final Configuration:")
    # print(conf.to_yaml())

    if args["rl_backend"] == "rllib":
        import ray
        ray.shutdown()
        ray.init(ignore_reinit_error=True)
        navsim.NavSimGymEnv.register_with_ray()
        result = ray.tune.run(
            ppo.PPOTrainer,
            config=conf,
            name=run_conf.run_id,
            resume=args.resume,
            local_dir=str(run_base_folder),
            stop={"episodes_total": run_conf.total_episodes},
            checkpoint_freq=run_conf.checkpoint_interval,
            checkpoint_at_end=True
        )
        best_checkpoint = result.get_last_checkpoint(
            metric="episode_reward_mean", mode="max"
        )
        # best_checkpoint = result.get_trial_checkpoints_paths(
        #    trial=result.get_best_trial("episode_reward_mean"),
        #    metric="episode_reward_mean", mode="max")

        print(best_checkpoint)
        trainer = ppo.PPOTrainer(config=conf)
        trainer.restore(best_checkpoint)
        model = trainer.get_policy().model
        print(type(model))
        print(model)
    else:
        # if resume is passed then read the args from saved conf instead and
        # then overwrite with the args passed
        # TODO: Optimize this with respect to above arg/config setting
        if args.resume and run_base_folder.is_dir():
            conf = ObjDict().load_from_json_file(
                str(run_base_folder / "conf.json"))

            # detect if physics lebvel has changed in resume
            if ("agent_car_physics" in non_default_args) and int(
                    conf["env_config"]["agent_car_physics"]) != int(
                args["agent_car_physics"]):
                # detect if clear memory was overridden from command prompt
                if "clear_memory" not in non_default_args:
                    conf["run_config"]["clear_memory"] = True
                else:
                    # clear_memory will be set from passed args below
                    pass

            for passed_arg in non_default_args:
                if passed_arg in conf["env_config"]:
                    conf["env_config"][passed_arg] = args[passed_arg]
                if passed_arg in conf["run_config"]:
                    conf["run_config"][passed_arg] = args[passed_arg]
                if passed_arg in conf:
                    conf[passed_arg] = args[passed_arg]


            for arg in ["seed", "memory_capacity", "batch_size",
                        "checkpoint_interval",
                        "train_interval", "agent_gpu_id"]:
                conf["run_config"][arg] = int(conf["run_config"][arg])
            for arg in ["discount", "tau", "expl_noise"]:
                conf["run_config"][arg] = float(conf["run_config"][arg])
            for arg in ["env_gpu_id", "seed", "timeout", "base_port",
                        "obs_mode", "obs_height", "obs_width",
                        "segmentation_mode", "episode_max_steps", "task",
                        "goal", "goal_distance", "traffic_vehicles",
                        "agent_car_physics"]:
                conf["env_config"][arg] = int(conf["env_config"][arg])
            for arg in ["reward_for_goal",
                        "reward_for_no_viable_path",
                        "reward_step_mul", "reward_collision_mul",
                        "reward_spl_delta_mul"]:
                conf["env_config"][arg] = float(conf["env_config"][arg])

        executor = Executor(resume=args["resume"],
                            conf=conf)
        executor.execute()
        print("training finished")

        executor.env_close()
        print("env closed")
        executor.files_close()
        print("files closed")


# For python debugger to directly run this script
if __name__ == "__main__":
    main()



import attr
import cattr
from . import ObjDict

@attr.s(auto_attribs=True)
class BaseConfig:
    def as_dict(self):
        return cattr.unstructure(self)

"""
@attr.s(auto_attribs=True)
class Config(BaseConfig):
    # Options: command line -> yaml_file -> defaults

    env_config: EnvConfig = attr.ib(factory=EnvConfig)
    run_config: EnvConfig = attr.ib(factory=EnvConfig)

    @staticmethod
    def from_args_dict(args):
        # create config dict from command line default values
        # If config file given in command line
        # if the config file doesnt exist then error and exit
        # else read config and overwrite default values
        # add command line values
        # let us load config file first
        config_file = args.config_file
        if config_file:
            try:
                conf = ObjDict().load_from_file(config_file)
            except FileNotFoundError as e:
                if args.config_file is not None:
                    abs_path = os.path.abspath(config_file)
                    raise OSError(f"Config file could not be found at {abs_path}.") from e
            if conf:
                print(f'Following configuration loaded from {config_file}')
                print(conf.to_yaml())
        else:
            print(f"Config file not specified in command line, continuing.")

        return True
        
        @attr.s(auto_attribs=True)
class RunConfig:

Configuration for running in general

    run_id: str = argparser.get_default("run_id")
    train = argparser.get_default("train")
    debug = argparser.get_default("num_envs")
    resume = argparser.get_default("resume")
    seed = argparser.get_default("force")
"""


class ReplayBuffer(object):

    def __init__(self, state_dimension, action_dimension, max_size=int(1e6)):
        self.max_size = max_size
        self.ptr = 0
        self.size = 0
        self.state = np.zeros((max_size, state_dimension))
        self.action = np.zeros((max_size, action_dimension))
        self.next_state = np.zeros((max_size, state_dimension))
        self.reward = np.zeros((max_size, 1))
        self.not_done = np.zeros((max_size, 1))
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    def add(self, state, action, next_state, reward, done):
        self.state[self.ptr] = state
        self.action[self.ptr] = action
        self.next_state[self.ptr] = next_state
        self.reward[self.ptr] = reward
        self.not_done[self.ptr] = 1. - done
        self.ptr = (self.ptr + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def sample(self, batch_size):
        ind = np.random.randint(0, self.size, size=batch_size)
        """
        return (
            torch.FloatTensor(self.state[ind]).to(self.device),
            torch.FloatTensor(self.action[ind]).to(self.device),
            torch.FloatTensor(self.next_state[ind]).to(self.device),
            torch.FloatTensor(self.reward[ind]).to(self.device),
            torch.FloatTensor(self.not_done[ind]).to(self.device)
        )
        """
        return (
            self.state[ind],
            self.action[ind],
            self.next_state[ind],
            self.reward[ind],
            self.not_done[ind]
        )



class ActorVector(torch.nn.Module):

    def __init__(self, state_dimension, action_dimension, max_action):
        super(ActorVector, self).__init__()
        self.l1 = torch.nn.Linear(state_dimension, 400)
        self.l2 = torch.nn.Linear(400, 300)
        self.l3 = torch.nn.Linear(300, action_dimension)
        self.max_action = max_action

    def forward(self, state):
        """

        :param state: state is a torch tensor
        :return:
        """
        # curframe = inspect.currentframe()
        # calframe = inspect.getouterframes(curframe, 2)
        # print('actor caller name:', calframe[1][3])
        # print('actor',state.shape)
        # print('actor',self.l1)
        # state = state[0]
        a = F.relu(self.l1(state))
        a = F.relu(self.l2(a))
        a = torch.tanh(self.l3(a))
        return self.max_action * a

class CriticVector(torch.nn.Module):

    def __init__(self, state_dimension, action_dimension):
        super(CriticVector, self).__init__()
        self.l1 = torch.nn.Linear(state_dimension + action_dimension, 400)
        self.l2 = torch.nn.Linear(400, 300)
        self.l3 = torch.nn.Linear(300, 1)

    def forward(self, state, action):
        """

        :param state: a torch Tensor
        :param action: a torch Tensor
        :return:
        """
        # print('critic',state.shape)
        # print('critic',self.l1)

        # state = state[0]
        q = F.relu(self.l1(torch.cat([state, action], 1)))
        q = F.relu(self.l2(q))
        q = self.l3(q)
        return q

class NavSimEnv:
    def __init__(self, conf):
        """
        conf: ObjDict having Environment Conf
        :param conf:
        """
        # filename: Optional[str] = None, obs_mode: int = 0, max_steps:int = 5):
        self.conf = conf

        self.observation_mode = self.conf['obs_mode']
        self.uenv = None
        self.genv = None

        self.env_open = False
        self.open()

    def open(self):

        if self.env_open:
            raise ValueError('Environment already open')
        else:
            log_folder = Path(self.conf['log_folder'])
            log_folder.mkdir(parents=True, exist_ok=True)

            engine_side_channel = EngineConfigurationChannel()
            environment_side_channel = EnvironmentParametersChannel()

            engine_side_channel.set_configuration_parameters(time_scale=10, quality_level=0)
            environment_side_channel.set_float_parameter("rewardForGoalCollision", self.conf['reward_for_goal'])
            environment_side_channel.set_float_parameter("rewardForExplorationPointCollision",
                                                         self.conf['reward_for_ep'])
            environment_side_channel.set_float_parameter("rewardForOtherCollision", self.conf['reward_for_other'])
            environment_side_channel.set_float_parameter("rewardForFallingOffMap",
                                                         self.conf['reward_for_falling_off_map'])
            environment_side_channel.set_float_parameter("rewardForEachStep", self.conf['reward_for_step'])
            environment_side_channel.set_float_parameter("segmentationMode", self.conf['segmentation_mode'])
            environment_side_channel.set_float_parameter("observationMode", self.conf['obs_mode'])
            environment_side_channel.set_float_parameter("episodeLength", self.conf['max_steps'])
            environment_side_channel.set_float_parameter("selectedTaskIndex", self.conf['task'])
            environment_side_channel.set_float_parameter("goalSelectionIndex", self.conf['goal'])
            environment_side_channel.set_float_parameter("agentCarPhysics", self.conf['agent_car_physics'])

            uenv_file_name = str(Path(self.conf['env_path']).resolve()) if self.conf['env_path'] else None
            self.uenv = UnityEnvironment(file_name=uenv_file_name,
                                         log_folder=str(log_folder.resolve()),
                                         seed=self.conf['seed'],
                                         timeout_wait=self.conf['timeout'],
                                         worker_id=self.conf['worker_id'],
                                         # base_port=self.conf['base_port'],
                                         no_graphics=False,
                                         side_channels=[engine_side_channel, environment_side_channel])

            self.genv = UnityToGymWrapper(self.uenv, False, False, True)
            # (Env, uint8_visual, flatten_branched, allow_multiple_obs)
            self.seed(self.conf['seed'])
            self.env_open = True

    def close(self):
        if self.env_open:
            self.env_open = False
            if self.uenv is None:
                print('uenv is None')
            else:
                self.uenv.close()
        else:
            raise ValueError('Environment not open')

    def info(self):
        print('-----------')
        print("Env Info")
        print('-----------')
        if self.genv.spec:
            print(self.genv.spec.id)
        print('Action Space:', self.action_space)
        print('Action sample:', self.action_space.sample())
        print('Action Space Shape:', self.action_space.shape)
        print('Action Space Low:', self.action_space.low)
        print('Action Space High:', self.action_space.high)
        print('Observation Mode:', self.observation_mode)
        print('Gym Observation Space:', self.genv.observation_space)
        print('Gym Observation Space Shape:', self.genv.observation_space.shape)
        print('Self Observation Space:', self.observation_space)
        print('Self Observation Space Shape:', self.observation_space.shape)
        print('Self Observation Space Shapes:', self.observation_space_shapes)
        print('Self Observation Space Types:', self.observation_space_types)
        print('Reward Range:', self.genv.reward_range)
        print('Metadata:', self.genv.metadata)

    def info_steps(self):
        print('Initial State:', self.reset())
        print('First Step State:', self.step(self.action_space.sample()))

    @property
    def action_space(self):
        return self.genv.action_space

    @property
    def action_space_shape(self):
        return self.action_space.shape

    @property
    def observation_space(self):
        """
        We can control here what observation space to return
        Observation Space is always returned as a tuple
        :return:
        """
        return self.genv.observation_space

    @property
    def observation_space_shapes(self):
        """
        Returns the dimensions of the observation space
        :return:
        """
        return [obs.shape for obs in self.observation_space.spaces]

    @property
    def observation_space_types(self):
        """
        Returns the dimensions of the observation space
        :return:
        """
        return [type(obs) for obs in self.observation_space.spaces]

    @property
    def gym_env(self):
        return self.genv

    @property
    def unity_env(self):
        return self.uenv

    def step(self, action):
        return self.genv.step(action)

    def reset(self):
        return self.genv.reset()

    def seed(self, seed=None):
        # TODO: This is broken in gym 0.10.9, fix it when they fix it
        # self.genv.action_space.seed(seed)
        # self.genv.seed(seed)
        pass
